#ifdef BASIN_HAVE_CFITSIO
#include "FITS_IO.h"

FITS_IO::FITS_IO (string fileName, Region* destination) {  

  _data = destination;
  _fileName = fileName;
  MPI_Comm_rank(BASIN_COMM, &_myRank);
  MPI_Comm_size(BASIN_COMM, &_nProcs);

  _doubleTypeKey = "double";
  _intTypeKey = "int";
  _stringTypeKey = "string";
  _floatTypeKey = "float";
  _complexTypeKey = "complex";
  //_vec2TypeKey = "vector2";
  //_vec3TypeKey = "vector3";
  _addNameKey = "_";

  _hduCount = 1;
  _iList = 0;
  
}

FITS_IO::~FITS_IO () { }

bool FITS_IO::OpenFile () {

  //Open the FITS
  int retVal = 1;
  int status = 0;
  
  if(_myRank == 0) {
    if(fits_open_file(&_filePtr, _fileName.c_str(), READONLY, &status)) {
      retVal = 0;
      fits_report_error(stderr,status);
    }
    
    if(fits_get_num_hdus(_filePtr, &_numHDUs, &status)) {
      retVal = 0;
      fits_report_error(stderr,status);
    }

    //cerr << "numHDUs = " << _numHDUs << endl;
  }
  

  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);
  MPI_Barrier(BASIN_COMM);

  return retVal;
}

bool FITS_IO::CloseFile () {

  //close the file
  int retVal = 1;
  int status = 0;
  if(_myRank == 0)
    if ( fits_close_file(_filePtr, &status) ) {
      retVal = 0;
      fits_report_error( stderr, status );
    }


  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);  
  MPI_Barrier(BASIN_COMM);
  

  return retVal;
}

unsigned long FITS_IO::NumElements() {
  long retVal;
  int status = 0;
  if(_myRank == 0) {
    fits_get_num_rows(_filePtr,&retVal,&status);
  }

  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);

  return retVal;
}

int FITS_IO::NumCols() {
  int retVal;
  int status = 0;
  if(_myRank == 0) {
    fits_get_num_cols(_filePtr,&retVal,&status);
  }

  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);

  return retVal;

} 


int FITS_IO::ReadCol(int iLo, int iHi, int col, int nrows, char* tType[], int typeCode) {
 
  int nElem = iHi - iLo + 1;

  
  //if(_myRank == 0)
  //cerr << "iHi = " << iHi << endl << "iLo = " << iLo << endl << "hduCount = " << _hduCount << endl;

  //long repeat;
  //int typeCode;
  //long width;
  int* anynul = 0;
  int nulval = 0;
  int status = 0;
  int retVal = 1;

  //MPI_Barrier(BASIN_COMM);
  //if(_myRank == 0)
  //cerr << "ReadCol" << endl;

  
  if(typeCode > 0 && _myRank == 0) {
    switch(typeCode) {   
    case TDOUBLE:
      //colType = _doubleTypeKey;
      _dArray.resize(nElem);      
      for(int i = iLo; i <= iHi; i++) 
	if(fits_read_col(_filePtr, typeCode, col, i+1, 1, 1,
			 &nulval, &_dArray[i-iLo], anynul, &status)) {
	  retVal = 0;
	  fits_report_error(stderr, status);
	}
      _newList -> set_attribute(tType[col-1], iLo, iHi, &_dArray[0]);   
      break;
	   
    case TFLOAT:
      _fArray.resize(nElem);      
      for(int i = iLo; i <= iHi; i++) 
	if(fits_read_col(_filePtr, typeCode, col, i+1, 1, 1,
			 &nulval, &_fArray[i-iLo], anynul, &status)) {
	  retVal = 0;
	  fits_report_error(stderr, status);
	}
      _newList -> set_attribute(tType[col-1], iLo, iHi, &_fArray[0]);
      break;

    case TLONG:
    case TSHORT:
      _iArray.resize(nElem);      
      for(int i = iLo; i <= iHi; i++) 
	if(fits_read_col(_filePtr, typeCode, col, i+1, 1, 1,
			 &nulval, &_iArray[i-iLo], anynul, &status)) {
	  retVal = 0;
	  fits_report_error(stderr, status);
	}
      _newList -> set_attribute(tType[col-1], iLo, iHi, &_iArray[0]);
      break;

    case TDBLCOMPLEX:
    case TCOMPLEX:
      _complexArray.resize(nElem);
      
      double dtemp;
      for(int i = iLo; i <= iHi; i++) {
	if(fits_read_col(_filePtr, typeCode, col, i+1, 1, 1, 
			 &nulval, &dtemp, anynul, &status)) {
	  retVal = 0;
	  fits_report_error(stderr, status);
	}
	  
	_complexArray[i-iLo].re = dtemp;
	
	if(fits_read_col(_filePtr, typeCode, col, i+1, 2, 1, 
			 &nulval, &dtemp, anynul, &status)) {
	  retVal = 0;
	  fits_report_error(stderr, status);
	}
	  
	_complexArray[i-iLo].im = dtemp;
	
      }

      _newList -> set_attribute(tType[col-1], iLo, iHi, &_complexArray[0]);
      break;
  
      
    }
    
  }
  
  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);   
  return retVal;
  
}

bool FITS_IO::NextDataset () {

  MPI_Barrier(BASIN_COMM);
  
  int retVal = 1;
  int status = 0;
  
  if(_myRank == 0) {
    if(_hduCount <= _numHDUs) {
      if(fits_movabs_hdu(_filePtr, _hduCount, &_hduType, &status)) {
	fits_report_error(stderr, status);
	retVal = 0;
      }
      //cerr << "_hduCount = " << _hduCount << endl << "hduType = " << _hduType << endl;;
      _hduCount++;
    }
    else
      retVal = 0;
  }

  MPI_Bcast(&_hduCount, 1, MPI_INT, 0, BASIN_COMM);
  MPI_Bcast(&_hduType, 1, MPI_INT, 0, BASIN_COMM);
  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);

  MPI_Barrier(BASIN_COMM);

  //if(_myRank == 0)
  //cerr << "retVal = " << retVal << endl;

  //if(_iList > 0)
  //return false;
  //return true;

  return retVal;
}

bool FITS_IO::MakeDataObjects () {
  
  int retVal = 1;
  
  //if(_myRank == 1)
  //cerr << "TESTING BEFORE FUCTION" << endl;
  
  if((_hduType == ASCII_TBL || _hduType == BINARY_TBL))
    {
      
      //if(_myRank ==  0)
      //cerr << "TESTING FUCTION" << endl;

      int nCols;
      unsigned long nRows;
      int retVal = 1;
      int nFound;
      int status = 0;

      vector<long> repeat;
      vector<int> typeCode;
      long width;
      int* anynul=0;
      int nulval = 0;
      string colName;
      
      nCols = NumCols();
      
      typeCode.resize(nCols);
      repeat.resize(nCols);

      int multiDim = 0;

      for(int i = 0; i < nCols && _myRank == 0 && multiDim == 0; i++) {
	if(fits_get_coltype(_filePtr,i+1,&typeCode[i],&repeat[i],&width,&status)) {
	  retVal = 0;
	  fits_report_error(stderr, status);
	}
	//cerr << typeCode[i] << endl;
	if(typeCode[i] != TSTRING)
	  if(repeat[i] > 1 || ((typeCode[i] == TCOMPLEX || typeCode[i] == TDBLCOMPLEX) && repeat[i] > 2))
	    multiDim = 1;
      }
      

      //else if(_hduType == ASCII_TBL)
      //if(_myRank == 0)
      //  cerr << "ASCII" << endl;

      MPI_Bcast(&multiDim, 1, MPI_INT, 0, BASIN_COMM);

      if(!multiDim) {
	nRows = NumElements();
	
	MPI_Bcast(&typeCode[0], nCols, MPI_INT, 0, BASIN_COMM);
	MPI_Bcast(&repeat[0], nCols, MPI_INT, 0, BASIN_COMM);

	MPI_Barrier(BASIN_COMM);
	//if(_myRank == 0)
	//cerr << "nRows = " << nRows << endl;
	
	//Make the list
	stringstream sstr;
	sstr << _iList;
	string index = sstr.str();
	_newList = _data -> add_list("RegionList" + index, nRows);
	vector<double> ll;
	ll.push_back(0);
	vector<double> ur;
	ur.push_back(1.0);
	vector<int> temp_nE;
	temp_nE.push_back(nRows);
	_newList->set_context(ll,ur,temp_nE);
	_iList++;
	
	
	MPI_Barrier(BASIN_COMM);
	//if(_myRank ==  0)
	//cerr << "TESTING FUCTION 2" << endl;
	
	char* tType[nCols];
	for(int i=0;i<nCols;i++)
	  tType[i]=(char *) malloc(FLEN_VALUE);
	
	if(_myRank == 0)
	  fits_read_keys_str(_filePtr, "TTYPE", 1, nCols, tType, &nFound, &status);

	MPI_Barrier(BASIN_COMM);
	//if(_myRank ==  0)
	//cerr << "TESTING FUCTION" << endl;
	
	for(int i = 0; i < nCols; i++) {
	  MPI_Bcast(tType[i], FLEN_VALUE, MPI_CHAR, 0, BASIN_COMM);
	  
	  MPI_Barrier(BASIN_COMM);

	  //if(_myRank == 0)
	  //if(fits_get_coltype(_filePtr,i+1,&typeCode,&repeat,&width,&status)) {
	  //  retVal = 0;
	  //  fits_report_error(stderr, status);
	  //}
	  
	  //if(_myRank == 0)
	  //cerr << "column " << i << ", typeCode = " << typeCode << ", repeat = " << repeat << endl;
	  
	  //MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);
	  //MPI_Bcast(&typeCode, 1, MPI_INT, 0, BASIN_COMM);
	  //MPI_Bcast(&repeat, 1, MPI_INT, 0, BASIN_COMM);
	  
	  //MPI_Barrier(BASIN_COMM);	  

	  if(typeCode[i] > 0) {
	    
	    //for(int rep = 0; rep < repeat; rep++) {
	    //stringstream sstr;
	    //sstr << rep;
	      
	    //if(repeat > 3 || (typeCode != TDOUBLE && repeat > 1))
	    //colName = tType[i] + _addNameKey + sstr.str();
	    //else
	    //colName = tType[i];
	      
	      
	    //int nameSize = colName.size();
	    //MPI_Bcast(&nameSize, 1, MPI_INT, 0, BASIN_COMM);
	    //if(_myRank != 0)
	    //colName.resize(nameSize);
	    //MPI_Bcast(&colName[0], nameSize, MPI_CHAR, 0, BASIN_COMM);	
	      
	      switch(typeCode[i]) {   
	      case TDOUBLE:
		_newList -> add_attribute(tType[i], double(0));
		break;
		
	      case TFLOAT:
		_newList -> add_attribute(tType[i], float(0));
		break;
		
	      case TLONG:
	      case TSHORT:
		_newList -> add_attribute(tType[i], int(0));
		break;
		
	      case TDBLCOMPLEX:
	      case TCOMPLEX:
		_newList -> add_attribute(tType[i], fftw_complex());
		break;
		
	      case TSTRING:
		if(_myRank == 0)
		  cerr << "FITS type TSTRING not supported" << endl;
		break;
		
	      }
	      //}
	  }
	  
	  MPI_Barrier(BASIN_COMM);
	  
	  //if(_myRank == 0)
	  //cerr << "TESTING AFTER ADD_ATTRIBUTE" << endl;

	  for(int proc = 0; proc < _nProcs; proc++) {
	    int iLo, iHi;
	    
	    _newList -> get_distribution_index(tType[i], proc, &iLo, &iHi);
	    if(!ReadCol(iLo, iHi, i+1, nRows, tType, typeCode[i]))
	      retVal = 0;
	  }
	}
	
	
	for(int i = 0; i < nCols; i++)
	  free(tType[i]);
	
	_newList = NULL;
      } //eo if(!multiDim)

      else
	if(_myRank == 0)
	  cerr << "Cannot read HDU number " << _hduCount - 1 << ", because it is a FITS binary table" << endl
	       << " with at least one column containing vector elements." << endl;
    } //eo if(hduType == ...      


  MPI_Barrier(BASIN_COMM);
  

  return retVal;

}

bool FITS_IO::NextGroup () {
  return false;
}
#endif
