#include "HDF5_IO.h"

HDF5_IO::HDF5_IO (string fileName, Region* destination) {
	_data = destination;
	_fileName = fileName;
	MPI_Comm_rank(BASIN_COMM, &_myRank);
}

HDF5_IO::~HDF5_IO () { }

bool HDF5_IO::OpenFile () {
	if (H5Fis_hdf5 (_fileName.c_str ()) > 0) {
		hid_t plist_id = H5Pcreate (H5P_FILE_ACCESS);
		H5Pset_fapl_mpiposix (plist_id, BASIN_COMM, (hbool_t) false);
		_fileID = H5Fopen (_fileName.c_str (), H5F_ACC_RDONLY, plist_id);
		H5Pclose (plist_id);
		return OpenRoot ();
	}

	if (_myRank == 0) cerr << _fileName << " is not a valid HDF5 file!" << endl;
	return false;
}

bool HDF5_IO::CloseFile () {
	if (!CloseRoot ()) {
		if (_myRank == 0) cerr << "Error during CloseRoot ()" << endl;
		return false;
	}
	if (H5Fclose (_fileID) < 0) {
		if (_myRank == 0) cerr << "Error during H5Fclose ()" << endl;
		return false;
	}
	return true;
}

bool HDF5_IO::OpenRoot () {
	_groupID.push (H5Gopen (_fileID, "/"));
	if (_groupID.top () < 0) return false;
	(*_data).set_name ("/");
	_currentRegion.push (_data);
	return true;
}

bool HDF5_IO::CloseRoot () {
	if (H5Gclose (_groupID.top ()) < 0) {
		if (_myRank == 0) cerr << "Error during H5Gclose ()" << endl;
		return false;
	}
	if (_groupID.empty ()) {
		if (_myRank == 0) cerr << "Error _groupID is empty, it shouldn't be!" << endl;
		return false;
	}
	_groupID.pop ();
	if (!_groupID.empty ()) {
		if (_myRank == 0) cerr << "Error _groupID isn't empty, it should be!" << endl;
		return false;
	}
	return true;
}

bool HDF5_IO::GetDatasets () {
	hsize_t numObjs;
	H5Gget_num_objs (_groupID.top (), &numObjs);
	vector <vector <int> *> dimPtrs;
	
	_datasetsLeft.clear ();
	for (unsigned int iObj = 0; iObj < numObjs; ++iObj) {
		char objName[101];
		int objType = H5Gget_objtype_by_idx (_groupID.top () ,iObj);
		H5Gget_objname_by_idx (_groupID.top (), iObj, objName, static_cast<size_t> (100));

		if (objType == 2) {
			hid_t datasetID = H5Dopen (_groupID.top (), objName);
			hid_t dataspaceID = H5Dget_space (datasetID);

			int dimensions = H5Sget_simple_extent_ndims (dataspaceID);
			hsize_t dimSize[dimensions], dimMaxSize[dimensions];
			H5Sget_simple_extent_dims (dataspaceID, dimSize, dimMaxSize);

			dimPtrs.push_back (new vector <int> ());
			for (unsigned int iTemp = 0; iTemp < dimensions; ++iTemp) {
				(*dimPtrs[dimPtrs.size () - 1]).push_back (dimSize[iTemp]);
			}
			
			H5Sclose (dataspaceID);
			H5Dclose (datasetID);
			
			_datasetsLeft.push_back (objName);
		}
	}

	for (unsigned int i = 0; i < dimPtrs.size (); ++i) {
		if (_datasetsSorted.empty ()) {
			_datasetsSorted.push_back (vector <int> (1, i));
		} else {
			bool test = false;
			for (vector<vector<int> >::iterator j = _datasetsSorted.begin (); j != _datasetsSorted.end (); j++) {
				if ((*dimPtrs[(*j)[0]]) == (*dimPtrs[i])) {
					(*j).push_back (i);
					test = true;
				}
			}

			if (test == false) {
				_datasetsSorted.push_back (vector <int> (1, i));
			}
		}
	}

	for (unsigned int i = 0; i < dimPtrs.size (); ++i) {
		delete dimPtrs[i];
	}

	return true;
}

bool HDF5_IO::GetGroups () {
	hsize_t numObjs;
	H5Gget_num_objs (_groupID.top (), &numObjs);
	_groupsLeft.push (stack<string> ());
	for (unsigned int iObj = 0; iObj < numObjs; ++iObj) {
		char objName[101];
		int objType = H5Gget_objtype_by_idx (_groupID.top (), iObj);
		H5Gget_objname_by_idx (_groupID.top (), iObj, objName, static_cast<size_t> (100));

		if (objType == 1) {
			_groupsLeft.top ().push (objName);
		}
	}
	return true;
}

bool HDF5_IO::NextDataset () {
	if (!_datasetsSorted.empty ()) {
		return true;
	}
	return false;
}

bool HDF5_IO::NextGroup () {
	if (!_groupsLeft.top ().empty ()) {
		return true;
	} else {
		_groupsLeft.pop ();
	}
	return false;
}

bool HDF5_IO::MakeDataObjects () {
	if (IsGrid (_datasetsSorted[_datasetsSorted.size () - 1])) {
		MakeGrid (_datasetsSorted[_datasetsSorted.size () - 1]);
	} else {
		MakeList (_datasetsSorted[_datasetsSorted.size () - 1]);
	}

	_datasetsSorted.pop_back ();
	return true;
}

bool HDF5_IO::MakeSubGroups () {
	_currentRegion.push ((*_currentRegion.top ()).add_child ());
	_groupID.push (H5Gopen (_groupID.top (), _groupsLeft.top ().top ().c_str ()));

	(*_currentRegion.top ()).set_name (_groupsLeft.top ().top ());
	ReadRegion ();

	H5Gclose (_groupID.top ());
	_groupID.pop ();
	_currentRegion.pop ();
	_groupsLeft.top ().pop ();
	return true;
}

bool HDF5_IO::MakeList (vector<int> attributeGrouping) {
	char objName[101];
	H5Iget_name(_groupID.top (), objName, static_cast<size_t> (100));
	ostringstream number;
	if (_datasetsSorted.size () > 1) {
		number << "#" << _datasetsSorted.size ();
	}
	string objNameCorrected = objName;
	int index = objNameCorrected.find ("#", 0);
	if (index) {
		objNameCorrected = objNameCorrected.substr (0, index);
	}
	string listName = objNameCorrected + number.str ();
	
	List* tempList = new List ();
	for (int j = 0; j < attributeGrouping.size (); ++j) {
		vector <Attribute*> tempAttrs = HDF5GetDataset (_groupID.top (), _datasetsLeft[attributeGrouping[j]].c_str ());
		unsigned int zeros = (int) log10 (tempAttrs.size ()) + 1;

		if (tempAttrs.size () == 1) {
			tempList->add_attribute (_datasetsLeft[attributeGrouping[j]], *(tempAttrs[0]));
			delete (tempAttrs[0]);
		} else {
			for (unsigned int k = 0; k < tempAttrs.size (); ++k) {
				unsigned int length = (int)log10 (k) + 1;
				ostringstream num;
				num << "#";
				while (length < zeros) {
					num << "0";
					++length;
				}
				num << k;
				tempList->add_attribute (_datasetsLeft[attributeGrouping[j]] + num.str (), (*tempAttrs[k]));
				delete (tempAttrs[k]);
			}
		}
	}
	_currentRegion.top ()->add_list (listName, tempList);
	return true;
}

bool HDF5_IO::MakeGrid (vector<int> attributeGrouping) {
	char objName[101];
	H5Iget_name(_groupID.top (), objName, static_cast<size_t> (100));
	ostringstream number;
	if (_datasetsSorted.size () > 1) {
		number << "#" << _datasetsSorted.size ();
	}
	string objNameCorrected = objName;
	int index = objNameCorrected.find ("#", 0);
	if (index) {
		objNameCorrected = objNameCorrected.substr (0, index);
	}
	string gridName = objNameCorrected + number.str ();
	
	hid_t datasetID = H5Dopen (_groupID.top (), _datasetsLeft[attributeGrouping[0]].c_str ());
	hid_t dataspaceID = H5Dget_space (datasetID);
	int dimensions = H5Sget_simple_extent_ndims (dataspaceID);
	hsize_t dimSize[dimensions], dimMaxSize[dimensions];
	H5Sget_simple_extent_dims (dataspaceID, dimSize, dimMaxSize);
	vector<int> dims;
	for (unsigned int iTemp = 0; iTemp < dimensions; ++iTemp) {
		dims.push_back (dimSize[iTemp]);
	}
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);

	Grid* tempGrid = new Grid (dims);
	for (int j = 0; j < attributeGrouping.size (); ++j) {
		vector <Attribute*> tempAttrs = HDF5GetDataset (_groupID.top (), _datasetsLeft[attributeGrouping[j]].c_str ());
		unsigned int zeros = (int) log10 (tempAttrs.size ()) + 1;

		if (tempAttrs.size () == 1) {
			(*tempGrid).add_attribute (_datasetsLeft[attributeGrouping[j]], *(tempAttrs[0]));
			delete tempAttrs[0];
		} else {
			for (unsigned int k = 0; k < tempAttrs.size (); ++k) {
				unsigned int length = (int) log10 (k) + 1;
				ostringstream num;
				num << "#";
				while (length < zeros) {
					num << "0";
					++length;
				}
				num << k;
				(*tempGrid).add_attribute (_datasetsLeft[attributeGrouping[j]] + num.str (), (*tempAttrs[k]));
				delete tempAttrs[k];
			}
		}
	}
	_currentRegion.top ()->add_grid (gridName, tempGrid);
	
	return true;
}

bool HDF5_IO::IsGrid (vector<int> attributeGrouping) {
	hid_t datasetID = H5Dopen (_groupID.top (), _datasetsLeft[attributeGrouping[0]].c_str ());
	hid_t dataspaceID = H5Dget_space (datasetID);

	if (H5Sget_simple_extent_ndims (dataspaceID) == 1) {
		H5Sclose (dataspaceID);
		H5Dclose (datasetID);
		return false;
	}

	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}

vector <Attribute*> HDF5_IO::HDF5GetDataset (hid_t parentID, const char * objName) {
	vector <Attribute*> attrPtr;

	/* Open current dataset */
	hid_t datasetID = H5Dopen (parentID, objName);

	/* Get dataspace's ID (no open),
	 * will show us dimensions of data */
	hid_t dataspaceID = H5Dget_space (datasetID);
	if (H5Sis_simple (dataspaceID) > 0) {
		/* Get number of dimensions */
		int dimensions = H5Sget_simple_extent_ndims (dataspaceID);
		if (dimensions > 0) {
			/* Get size of each dimension */
			hsize_t dimSize[dimensions], dimMaxSize[dimensions];
			H5Sget_simple_extent_dims (dataspaceID, dimSize, dimMaxSize);
			/* Get total number of elements */
			unsigned long int elements = H5Sget_simple_extent_npoints (dataspaceID);

			/* Get data type */
			hid_t datatypeID = H5Dget_type (datasetID);
			H5T_class_t dataclass = H5Tget_class(datatypeID);

			if (dataclass == H5T_ARRAY) {
				hid_t dataType = H5Tget_native_type (datatypeID, H5T_DIR_ASCEND);
				attrPtr = HDF5Array (elements, datasetID, dataType, dimensions, dimSize);
				H5Tclose (dataType);
			} else {
			if (dataclass == H5T_INTEGER) {
				hid_t dataType = H5Tget_native_type (datatypeID, H5T_DIR_ASCEND);
				attrPtr = HDF5Integer (elements, datasetID, dataType, dimensions, dimSize);
				H5Tclose (dataType);
			} else {
			if (dataclass == H5T_FLOAT) {
				hid_t dataType = H5Tget_native_type (datatypeID, H5T_DIR_ASCEND);
				attrPtr = HDF5Float (elements, datasetID, dataType, dimensions, dimSize);
				H5Tclose (dataType);
			} else {
			if (dataclass == H5T_STRING) {
				if (_myRank == 0) cerr << "*WARNING* " << objName << ": Strings are not implemented in BASIN" << endl;
			} else {
			if (dataclass == H5T_COMPOUND) {
				hid_t dataType = H5Tget_native_type (datatypeID, H5T_DIR_ASCEND);
				attrPtr = HDF5Compound (elements, datasetID, dataType, dimensions, dimSize);
				H5Tclose (dataType);
			} else {
			if (dataclass == H5T_ENUM) {
				hid_t dataType = H5Tget_native_type (datatypeID, H5T_DIR_ASCEND);
				attrPtr = HDF5Enum (elements, datasetID, dataType, dimensions, dimSize);
				H5Tclose (dataType);
			} else {
			if (dataclass == H5T_REFERENCE) {
				if (_myRank == 0) cerr << "*WARNING* " << objName << ": References are not implemented in BASIN" << endl;
			} else {
			if (dataclass == H5T_BITFIELD) {
				if (_myRank == 0) cerr << "*WARNING* " << objName << ": Bitfields are not implemented in BASIN" << endl;
			} else {
			if (dataclass == H5T_OPAQUE) {
				if (_myRank == 0) cerr << "*WARNING* " << objName << ": Opaques are not implemented in BASIN" << endl;
			} else {
			if (dataclass == H5T_VLEN) {
				if (_myRank == 0) cerr << "*WARNING* " << objName << ": Variable Lengths are not implemented in BASIN" << endl;
			} else {
			}}}}}}}}}}
			H5Tclose (datatypeID);
		} else {
			if (_myRank == 0) cerr << objName << ": BASIN doesn't allow " << dimensions << " dimensional data!" << endl;
		}
	} else {
		if (_myRank == 0) cerr << objName << ": Dataspace is complex, doing nothing!" << endl;
	}

	/* Close current dataset & dataspace*/
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);

	return attrPtr;
}

vector <Attribute*> HDF5_IO::HDF5Integer (unsigned long int elements, hid_t datasetID, hid_t dataType, hsize_t dimensions, hsize_t *dimSize) {
	H5T_class_t dataclass = H5Tget_class(dataType);
	hid_t dataTypeID, dataTypeNative;
	if (dataclass == H5T_COMPOUND) {
		dataTypeID = H5Tget_member_type(dataType, 0);
		dataTypeNative = H5Tget_native_type (dataTypeID, H5T_DIR_ASCEND);
	}

	/* Vectorize the dimSize, and make a "dimSize" vector of 0's
	 * So that you have two vectors that describe the boundaries
	 * of *data[] that you wish to place in the attribute */
	vector <int> dimVec (dimensions);
	vector <int> dimUpper (dimensions);
	vector <int> dimLower (dimensions);
	for (int i = 0; i < dimensions; ++i) {
		dimVec[i] = dimSize[i];
		dimUpper[i] = dimSize[i] - 1;
		dimLower[i] = 0;
	}
	vector <Attribute*> attrPtrs;

	unsigned int intBytes;
	hid_t datatypeID;
	if (dataType == -1) {
		datatypeID = H5Dget_type (datasetID);
		intBytes = H5Tget_size (datatypeID);
	} else {
		intBytes = H5Tget_size (dataType);
	}
	
	/* Initializes BASIN attribute, and retrieves the breakdown of elements of each process */
	int dummy = 0;
	Attribute * attr = new Attribute (dimVec, dummy);
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;

	hid_t spaceIDFile = H5Dget_space (datasetID);
	hid_t spaceIDMem;

	if (dimensions > 1) {
		hsize_t* beginCoords = new hsize_t[dimensions];
		hsize_t* finalCoords = new hsize_t[dimensions];
		hsize_t* countCoords = new hsize_t[dimensions];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dimensions - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dimSize[iDim]) % dimSize[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dimSize[iDim]) % dimSize[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			if (countCoords[iDim] < 1) {
				cerr << "Empty Dimension!" << endl;
				//return attrPtrs;
			}
			mod *= dimSize[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dimensions, countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dimensions == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		if (countCoords[0] < 1) {
			cerr << "Empty Dimension (single)!" << endl;
			//return attrPtrs;
		}
		H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}
	H5Sselect_all (spaceIDMem);

	void* data;
	if (dataTypeNative == H5T_NATIVE_SHORT || intBytes == 2) {
		data = new short[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_INT || intBytes == 4) {
		data = new int[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_LONG || intBytes == 8) {
		data = new long[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_LLONG || intBytes == 16) {
		data = new long long[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_USHORT || intBytes == 2) {
		data = new unsigned short[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_UINT || intBytes == 4) {
		data = new unsigned int[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_ULONG || intBytes == 8) {
		data = new unsigned long[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_ULLONG || intBytes == 16) {
		data = new unsigned long long[procElements];
	} else {
		data = new int[procElements];
	}}}}}}}}

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);

	/* Actually get the data from file */
	H5Dread (datasetID, dataType, spaceIDMem, spaceIDFile, plistID, data);

	H5Pclose (plistID);


	if (dataTypeNative == H5T_NATIVE_SHORT || intBytes == 2) {
		int* tempData = new int[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<int> (((short*) data)[iElem]);
		}
		delete [] (short*)data;
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] tempData;
	} else {
	if (dataTypeNative == H5T_NATIVE_INT || intBytes == 4) {
		(*attr).put (procInds[0], procInds[1], (int*) data);
		delete [] (int*)data;
	} else {
	if (dataTypeNative == H5T_NATIVE_LONG || intBytes == 8) {
		int* tempData = new int[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<int> (((long*) data)[iElem]);
		}
		delete [] (long*)data;
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] tempData;
	} else {
	if (dataTypeNative == H5T_NATIVE_LLONG || intBytes == 16) {
		int* tempData = new int[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<int> (((long long*) data)[iElem]);
		}
		delete [] (long long*)data;
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] tempData;
	} else {
	if (dataTypeNative == H5T_NATIVE_USHORT || intBytes == 2) {
		int* tempData = new int[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<int> (((unsigned short*) data)[iElem]);
		}
		delete [] (unsigned short*)data;
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] tempData;
	} else {
	if (dataTypeNative == H5T_NATIVE_UINT || intBytes == 4) {
		int* tempData = new int[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<int> (((unsigned int*) data)[iElem]);
		}
		delete [] (unsigned int*)data;
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] tempData;
	} else {
	if (dataTypeNative == H5T_NATIVE_ULONG || intBytes == 8) {
		int* tempData = new int[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<int> (((unsigned long*) data)[iElem]);
		}
		delete [] (unsigned long*)data;
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] tempData;
	} else {
	if (dataTypeNative == H5T_NATIVE_ULLONG || intBytes == 16) {
		int* tempData = new int[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<int> (((unsigned long long*) data)[iElem]);
		}
		delete [] (unsigned long long*)data;
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] tempData;
	} else {
		(*attr).put (procInds[0], procInds[1], (int*) data);
		delete [] (int*)data;
	}}}}}}}}

	/* Close up dataspace */
	H5Sclose (spaceIDMem);
	H5Sclose (spaceIDFile);
	if (dataclass == H5T_COMPOUND) {
		H5Tclose (dataTypeID);
		H5Tclose (dataTypeNative);
	}
	
	attrPtrs.push_back (attr);
	return attrPtrs;
}

vector <Attribute*> HDF5_IO::HDF5Float (unsigned long int elements, hid_t datasetID, hid_t dataType, hsize_t dimensions, hsize_t *dimSize) {
	H5T_class_t dataclass = H5Tget_class(dataType);
	hid_t dataTypeID, dataTypeNative;
	if (dataclass == H5T_COMPOUND) {
		dataTypeID = H5Tget_member_type(dataType, 0);
		dataTypeNative = H5Tget_native_type (dataTypeID, H5T_DIR_ASCEND);
	}
	/* Vectorize the dimSize, and make a "dimSize" vector of 0's
	 * So that you have two vectors that describe the boundaries
	 * of *data[] that you wish to place in the attribute */
	vector <int> dimVec (dimensions);
	vector <int> dimUpper (dimensions);
	vector <int> dimLower (dimensions);
	for (int i = 0; i < dimensions; ++i) {
		dimVec[i] = dimSize[i];
		dimUpper[i] = dimSize[i] - 1;
		dimLower[i] = 0;
	}

	vector <Attribute*> attrPtrs;

	unsigned int floatBytes;
	hid_t datatypeID;
	if (dataType == -1) {
		datatypeID = H5Dget_type (datasetID);
		floatBytes = H5Tget_size (datatypeID);
	} else {
		floatBytes = H5Tget_size (dataType);
	}

	/* Initializes BASIN attribute, and retrieves the breakdown of elements of each process */
	Attribute* attr;
	if (dataTypeNative == H5T_NATIVE_FLOAT || floatBytes == H5Tget_size (H5T_NATIVE_FLOAT)) {
		attr = new Attribute (dimVec, (float) 0.0);
	} else {
	if (dataTypeNative == H5T_NATIVE_DOUBLE  || floatBytes == H5Tget_size (H5T_NATIVE_DOUBLE)) {
		attr = new Attribute (dimVec, (double) 0.0);
	} else {
	if (dataTypeNative == H5T_NATIVE_LDOUBLE  || floatBytes == H5Tget_size (H5T_NATIVE_LDOUBLE)) {
		attr = new Attribute (dimVec, (double) 0.0);
	} else {
		return vector<Attribute*> ();
	}}}
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;

	hid_t spaceIDFile = H5Dget_space (datasetID);
	hid_t spaceIDMem;

	if (dimensions > 1) {
		hsize_t* beginCoords = new hsize_t[dimensions];
		hsize_t* finalCoords = new hsize_t[dimensions];
		hsize_t* countCoords = new hsize_t[dimensions];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dimensions - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dimSize[iDim]) % dimSize[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dimSize[iDim]) % dimSize[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			if (countCoords[iDim] < 1) {
				cerr << "Empty Dimension!" << endl;
				//return attrPtrs;
			}
			mod *= dimSize[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dimensions, countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dimensions == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		if (countCoords[0] < 1) {
			cerr << "Empty Dimension (single)!" << endl;
			//return attrPtrs;
		}
		H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}
	H5Sselect_all (spaceIDMem);

	void* data;
	if (dataTypeNative == H5T_NATIVE_FLOAT || floatBytes == H5Tget_size (H5T_NATIVE_FLOAT)) {
		data = new float[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_DOUBLE || floatBytes == H5Tget_size (H5T_NATIVE_DOUBLE)) {
		data = new double[procElements];
	} else {
	if (dataTypeNative == H5T_NATIVE_LDOUBLE || floatBytes == H5Tget_size (H5T_NATIVE_LDOUBLE)) {
		data = new long double[procElements];
	} else {
		return vector<Attribute*> ();
	}}}

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);

	/* Actually get the data from file */
	H5Dread (datasetID, dataType, spaceIDMem, spaceIDFile, plistID, data);

	H5Pclose (plistID);

	/* Commit data to the attribute */

	if (dataTypeNative == H5T_NATIVE_FLOAT || floatBytes == H5Tget_size (H5T_NATIVE_FLOAT)) {
		(*attr).put (procInds[0], procInds[1], (float*) data);
		delete [] (double*)data;
	} else {
	if (dataTypeNative == H5T_NATIVE_DOUBLE || floatBytes == H5Tget_size (H5T_NATIVE_DOUBLE)) {
		(*attr).put (procInds[0], procInds[1], (double*) data);
		delete [] (double*)data;
	} else {
	if (dataTypeNative == H5T_NATIVE_LDOUBLE || floatBytes == H5Tget_size (H5T_NATIVE_LDOUBLE)) {
		double* tempData = new double[procElements];
		for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
			tempData[iElem] = static_cast<double> (((long double*) data)[iElem]);
		}
		(*attr).put (procInds[0], procInds[1], tempData);
		delete [] (long double*)data;
		delete [] tempData;
	} else {
		return vector<Attribute*> ();
	}}}

	/* Close up dataspace */
	H5Sclose (spaceIDMem);
	H5Sclose (spaceIDFile);
	if (dataclass == H5T_COMPOUND) {
		H5Tclose (dataTypeID);
		H5Tclose (dataTypeNative);
	}

	attrPtrs.push_back (attr);
	return attrPtrs;
}

/* Enumerations aren't truly supported in BASIN, however I will convert them to integers which are supported. */
vector <Attribute*> HDF5_IO::HDF5Enum (unsigned long int elements, hid_t datasetID, hid_t dataType, hsize_t dimensions, hsize_t *dimSize) {
	H5T_class_t dataclass = H5Tget_class(dataType);
	hid_t dataTypeID, dataTypeNative;
	if (dataclass == H5T_COMPOUND) {
		dataTypeID = H5Tget_member_type(dataType, 0);
		dataTypeNative = H5Tget_native_type (dataTypeID, H5T_DIR_ASCEND);
	}
	vector <Attribute*> attrPtrs;

	/* Vectorize the dimSize, and make a "dimSize" vector of 0's
	 * So that you have two vectors that describe the boundaries
	 * of *data[] that you wish to place in the attribute */
	vector <int> dimVec (dimensions);
	vector <int> dimUpper (dimensions);
	vector <int> dimLower (dimensions);
	for (int i = 0; i < dimensions; ++i) {
		dimVec[i] = dimSize[i];
		dimUpper[i] = dimSize[i] - 1;
		dimLower[i] = 0;
	}

	/* Initializes BASIN attribute, and retrieves the breakdown of elements of each process */
	int dummy = 0;
	Attribute* attr = new Attribute (dimVec, dummy);
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	
	hid_t spaceIDFile = H5Dget_space (datasetID);
	hid_t spaceIDMem;

	if (dimensions > 1) {
		hsize_t* beginCoords = new hsize_t[dimensions];
		hsize_t* finalCoords = new hsize_t[dimensions];
		hsize_t* countCoords = new hsize_t[dimensions];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dimensions - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dimSize[iDim]) % dimSize[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dimSize[iDim]) % dimSize[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			if (countCoords[iDim] < 1) {
				cerr << "Empty Dimension!" << endl;
				//return attrPtrs;
			}
			mod *= dimSize[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dimensions, countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dimensions == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);
		if (countCoords[0] < 1) {
			cerr << "Empty Dimension (single)!" << endl;
			//return attrPtrs;
		}

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}
	H5Sselect_all (spaceIDMem);

	int* data = new int[procElements];

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);
	
	/* Actually get the data from file */
	H5Dread (datasetID, dataType, spaceIDMem, spaceIDFile, plistID, data);

	H5Pclose (plistID);

	/* Commit data to the attribute */
	(*attr).put (procInds[0], procInds[1], data);

	delete [] data;

	/* Close up dataspace */
	H5Sclose (spaceIDMem);
	H5Sclose (spaceIDFile);
	if (dataclass == H5T_COMPOUND) {
		H5Tclose (dataTypeID);
		H5Tclose (dataTypeNative);
	}
	
	attrPtrs.push_back (attr);
	return attrPtrs;
}

vector <Attribute*> HDF5_IO::HDF5Array (unsigned long int elements, hid_t datasetID, hid_t datatypeID, hsize_t dimensions, hsize_t *dimSize) {
	H5T_class_t dataclass = H5Tget_class(datatypeID);
	hid_t dataTypeID, datatypeIDNative;
	if (dataclass == H5T_COMPOUND) {
		dataTypeID = H5Tget_member_type(datatypeID, 0);
		datatypeIDNative = H5Tget_native_type (dataTypeID, H5T_DIR_ASCEND);
	}
	vector <Attribute*> attrPtrs;

	int arrayDims;
	hsize_t *arrayDimSizes;
	if (dataclass == H5T_COMPOUND) {
		arrayDims = H5Tget_array_ndims (dataTypeID);
		arrayDimSizes = new hsize_t[arrayDims];
		H5Tget_array_dims (dataTypeID, arrayDimSizes, NULL);
	} else {
		arrayDims = H5Tget_array_ndims (datatypeID);
		arrayDimSizes = new hsize_t[arrayDims];
		H5Tget_array_dims (datatypeID, arrayDimSizes, NULL);
	}
	unsigned int arrayElements = 1;
	for (unsigned int iDim = 0; iDim < arrayDims; ++iDim) {
		arrayElements *= arrayDimSizes[iDim];
	}

	/* Vectorize the dimSize, and make a "dimSize" vector of 0's
	 * So that you have two vectors that describe the boundaries
	 * of *data[] that you wish to place in the attribute */
	/* Note that these dimensions have nothing to do with the internal
	 * array, just the set up of arrays */
	vector <int> dimVec (dimensions);
	vector <int> dimUpper (dimensions);
	vector <int> dimLower (dimensions);
	for (int i = 0; i < dimensions; ++i) {
		dimVec[i] = dimSize[i];
		dimUpper[i] = dimSize[i] - 1;
		dimLower[i] = 0;
	}

	/* Handle special cases of arrays for BASIN */
	bool isVec2 = false;
	bool isVec3 = false;
	bool isComplex = false;
	int numAttributes = H5Aget_num_attrs (datasetID);
	for (unsigned int iAttr = 0; iAttr < numAttributes; ++iAttr) {
		hid_t attributeID = H5Aopen_idx (datasetID, iAttr);

		char attributeName[21];
		H5Aget_name (attributeID, static_cast<size_t> (20), attributeName);

		if ((arrayElements == 2 && arrayDims == 1) && (static_cast<string> (attributeName) == "BASIN_VECTOR2")) {
#ifdef VECTOR23
			isVec2 = true;
#endif
		} else {
		if ((arrayElements == 3 && arrayDims == 1) && (static_cast<string> (attributeName) == "BASIN_VECTOR3")) {
#ifdef VECTOR23
			isVec3 = true;
#endif
		} else {
		if ((arrayElements == 2 && arrayDims == 1) && (static_cast<string> (attributeName) == "BASIN_COMPLEX")) {
			isComplex = true;
		}}}

		H5Aclose (attributeID);
	}

	int bytes = H5Tget_size (datatypeID) / arrayElements;

	if (H5Tdetect_class(datatypeID, H5T_INTEGER)) {
		for (unsigned int i = 0; i < arrayElements; ++i) {
			attrPtrs.push_back (new Attribute (dimVec, (int) 0));
		}

		vector <int> procInds = (*(attrPtrs[0])).get_gen_distribution_index(_myRank);
		unsigned int procElements = procInds[1] - procInds[0] + 1;

		hid_t spaceIDFile = H5Dget_space (datasetID);
		hid_t spaceIDMem;

		if (dimensions > 1) {
			hsize_t* beginCoords = new hsize_t[dimensions];
			hsize_t* finalCoords = new hsize_t[dimensions];
			hsize_t* countCoords = new hsize_t[dimensions];

			/* Converts flattened array into n rank coordinates */
			int mod = 1;
			for (int iDim = dimensions - 1; iDim >= 0; --iDim) {
				beginCoords[iDim] = ((procInds[0] / mod) + dimSize[iDim]) % dimSize[iDim];
				finalCoords[iDim] = ((procInds[1] / mod) + dimSize[iDim]) % dimSize[iDim];
				countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
				if (countCoords[iDim] < 1) {
					cerr << "Empty Dimension! (bad)" << endl;
					return vector<Attribute*> ();
				}
				mod *= dimSize[iDim];
			}

			/* Make hyperslab */
			H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

			/* No longer needed, only for setting up hyperslab */
			delete [] beginCoords;
			delete [] finalCoords;

			/* Where the data from reading the hdf5 file will go */
			spaceIDMem = H5Screate_simple (dimensions, countCoords, NULL);
			delete [] countCoords;
		} else {
		if (dimensions == 1) {
			/* Make hyperslab */
			hsize_t beginCoords[1] = {procInds[0]};
			hsize_t countCoords[1] = {procElements};
			if (countCoords[0] < 1) {
				cerr << "Empty Dimension (single) (bad)!" << endl;
				return vector<Attribute*> ();
			}
			H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

			/* Where the data from reading the hdf5 file will go */
			spaceIDMem = H5Screate_simple (1, countCoords, NULL);
		}}
		H5Sselect_all (spaceIDMem);

		void* data;
		if (datatypeIDNative == H5T_NATIVE_SHORT || bytes == 2) {
			data = new short[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_INT || bytes == 4) {
			data = new int[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_LONG || bytes == 8) {
			data = new long[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_LLONG || bytes == 16) {
			data = new long long[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_USHORT || bytes == 2) {
			data = new unsigned short[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_UINT || bytes == 4) {
			data = new unsigned int[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_ULONG || bytes == 8) {
			data = new unsigned long[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_ULLONG || bytes == 16) {
			data = new unsigned long long[procElements * arrayElements];
		} else {
			if (_myRank == 0) cerr << "Warning: Array - unrecognized integer type" << endl;
			return vector<Attribute*> ();
		}}}}}}}}

		hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
		H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);

		/* Actually get the data from file */
		H5Dread (datasetID, datatypeID, spaceIDMem, spaceIDFile, plistID, data);

		H5Pclose (plistID);
		H5Sclose (spaceIDMem);
		H5Sclose (spaceIDFile);

		/* Putting data into attribute, then deleting data */
		if (datatypeIDNative == H5T_NATIVE_SHORT || bytes == 2) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((short*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (short*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_INT || bytes == 4) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((int*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (int*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_LONG || bytes == 8) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((long*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (long*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_LLONG || bytes == 16) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((long long*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (long long*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_USHORT || bytes == 2) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((unsigned short*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (unsigned short*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_UINT || bytes == 4) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((unsigned int*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (unsigned int*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_ULONG || bytes == 8) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((unsigned long*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (unsigned long*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_ULLONG || bytes == 16) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				int* tempData = new int[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = static_cast<int> (((unsigned long long*) data)[iElem * (iArrayElem + 1)]);
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (unsigned long long*) data;
		} else {
			return vector<Attribute*> ();
		}}}}}}}}
	} else {
	if (H5Tdetect_class(datatypeID, H5T_FLOAT)) {
		if (datatypeIDNative == H5T_NATIVE_FLOAT || bytes == H5Tget_size (H5T_NATIVE_FLOAT)) {
			for (unsigned int i = 0; i < arrayElements; ++i) {
				attrPtrs.push_back (new Attribute (dimVec, (float) 0.0));
			}
		} else {
		if (datatypeIDNative == H5T_NATIVE_DOUBLE || bytes == H5Tget_size (H5T_NATIVE_DOUBLE)) {
			if (isVec2) {
#ifdef VECTOR23
			  attrPtrs = vector<Attribute*> (1, new Attribute (dimVec, (Vector2) 0.0));
#endif			 
			} else {
			if (isVec3) {
#ifdef VECTOR23
				attrPtrs = vector<Attribute*> (1, new Attribute (dimVec, (Vector3) 0.0));
#endif
			} else {
			if (isComplex) {
				attrPtrs = vector<Attribute*> (1, new Attribute (dimVec, fftw_complex ()));
			} else {
				for (unsigned int i = 0; i < arrayElements; ++i) {
					attrPtrs.push_back (new Attribute (dimVec, (double) 0.0));
				}
			}}}
		} else {
		if (datatypeIDNative == H5T_NATIVE_LDOUBLE || bytes == H5Tget_size (H5T_NATIVE_LDOUBLE)) {
			for (unsigned int i = 0; i < arrayElements; ++i) {
				attrPtrs.push_back (new Attribute (dimVec, (double) 0.0));
			}
		} else {
			return vector<Attribute*> ();
		}}}

		vector <int> procInds = (*(attrPtrs[0])).get_gen_distribution_index(_myRank);
		unsigned int procElements = procInds[1] - procInds[0] + 1;

		hid_t spaceIDFile = H5Dget_space (datasetID);
		hid_t spaceIDMem;

		if (dimensions > 1) {
			hsize_t* beginCoords = new hsize_t[dimensions];
			hsize_t* finalCoords = new hsize_t[dimensions];
			hsize_t* countCoords = new hsize_t[dimensions];

			/* Converts flattened array into n rank coordinates */
			int mod = 1;
			for (int iDim = dimensions - 1; iDim >= 0; --iDim) {
				beginCoords[iDim] = ((procInds[0] / mod) + dimSize[iDim]) % dimSize[iDim];
				finalCoords[iDim] = ((procInds[1] / mod) + dimSize[iDim]) % dimSize[iDim];
				countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
				if (countCoords[iDim] < 1) {
					cerr << "Empty Dimension (bad)!" << endl;
					return vector<Attribute*> ();
				}
				mod *= dimSize[iDim];
			}

			/* Make hyperslab */
			H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

			/* No longer needed, only for setting up hyperslab */
			delete [] beginCoords;
			delete [] finalCoords;

			/* Where the data from reading the hdf5 file will go */
			spaceIDMem = H5Screate_simple (dimensions, countCoords, NULL);
			delete [] countCoords;
		} else {
		if (dimensions == 1) {
			/* Make hyperslab */
			hsize_t beginCoords[1] = {procInds[0]};
			hsize_t countCoords[1] = {procElements};
			if (countCoords[0] < 1) {
				cerr << "Empty Dimension (single) (bad)!" << endl;
				return vector<Attribute*> ();
			}
			H5Sselect_hyperslab (spaceIDFile, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

			/* Where the data from reading the hdf5 file will go */
			spaceIDMem = H5Screate_simple (1, countCoords, NULL);
		}}
		H5Sselect_all (spaceIDMem);

		void* data;
		if (datatypeIDNative == H5T_NATIVE_FLOAT || bytes == H5Tget_size (H5T_NATIVE_FLOAT)) {
			data = new float[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_DOUBLE || bytes == H5Tget_size (H5T_NATIVE_DOUBLE)) {
			data = new double[procElements * arrayElements];
		} else {
		if (datatypeIDNative == H5T_NATIVE_LDOUBLE || bytes == H5Tget_size (H5T_NATIVE_LDOUBLE)) {
			data = new long double[procElements * arrayElements];
		} else {
			if (_myRank == 0) cerr << "Warning: Array - unrecognized floating point type" << endl;
			return vector<Attribute*> ();
		}}}

		hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
		H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);

		/* Actually get the data from file */
		H5Dread (datasetID, datatypeID, spaceIDMem, spaceIDFile, plistID, data);

		H5Pclose (plistID);
		H5Sclose (spaceIDMem);
		H5Sclose (spaceIDFile);

		/* Putting data into attribute, then deleting data */
		if (datatypeIDNative == H5T_NATIVE_FLOAT || bytes == H5Tget_size (H5T_NATIVE_FLOAT)) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				float* tempData = new float[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = ((float*) data)[iElem * (iArrayElem + 1)];
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (float*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_DOUBLE || bytes == H5Tget_size (H5T_NATIVE_DOUBLE)) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				if (isVec2) {
#ifdef VECTOR23
					Vector2* tempData = new Vector2[procElements];
					for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
						tempData[iElem] = ((Vector2*) data)[iElem * (iArrayElem + 1)];
					}
					(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
					delete [] tempData;
#endif
				} else {
				if (isVec3) {
#ifdef VECTOR23
					Vector3* tempData = new Vector3[procElements];
					for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
						tempData[iElem] = ((Vector3*) data)[iElem * (iArrayElem + 1)];
					}
					(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
					delete [] tempData;
#endif
				} else {
				if (isComplex) {
					fftw_complex* tempData = new fftw_complex[procElements];
					for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
						tempData[iElem] = ((fftw_complex*) data)[iElem * (iArrayElem + 1)];
					}
					(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
					delete [] tempData;
				} else {
					double* tempData = new double[procElements];
					for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
						tempData[iElem] = ((double*) data)[iElem * (iArrayElem + 1)];
					}
					(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
					delete [] tempData;
				}}}
			}
			delete [] (double*) data;
		} else {
		if (datatypeIDNative == H5T_NATIVE_LDOUBLE || bytes == H5Tget_size (H5T_NATIVE_LDOUBLE)) {
			for (unsigned long int iArrayElem = 0; iArrayElem < arrayElements; ++iArrayElem) {
				double* tempData = new double[procElements];
				for (unsigned int iElem = 0; iElem < procElements; ++iElem) {
					tempData[iElem] = ((long double*) data)[iElem * (iArrayElem + 1)];
				}
				(*(attrPtrs[iArrayElem])).put (dimLower, dimUpper, tempData);
				delete [] tempData;
			}
			delete [] (long double*) data;
		}}}
	}}
	delete [] arrayDimSizes;
	if (dataclass == H5T_COMPOUND) {
		H5Tclose (dataTypeID);
		H5Tclose (datatypeIDNative);
	}
	return attrPtrs;
}

vector <Attribute*> HDF5_IO::HDF5Compound (unsigned long int elements, hid_t datasetID, hid_t dataType, hsize_t dimensions, hsize_t *dimSize) {
	/* Vectorize the dimSize, and make a "dimSize" vector of 0's
	 * So that you have two vectors that describe the boundaries
	 * of *data[] that you wish to place in the attribute */
	vector <int> dimUpper (dimensions);
	for (int i = 0; i < dimensions; ++i) {
		dimUpper[i] = dimSize[i];
	}

	vector <Attribute*> attrPtrs;

	hid_t dataTypeID = H5Dget_type (datasetID);
	unsigned long int compoundMembers = H5Tget_nmembers (dataTypeID);

	unsigned long int iMember;
	for (iMember = 0; iMember < compoundMembers; ++iMember) {
		H5T_class_t tempDataclass = H5Tget_member_class(dataTypeID, iMember);
		hid_t memberType = H5Tget_member_type (dataTypeID, iMember);
		char *memberName = H5Tget_member_name (dataTypeID, iMember);
		hid_t tempType;

		tempType = H5Tcreate (H5T_COMPOUND, H5Tget_size (memberType));
		H5Tinsert (tempType, memberName, H5Tget_offset(memberType), memberType);

		/* Once the class has been determined,
		 * send it to existing print functions
		 * to be handled */
		if (tempDataclass == H5T_ARRAY) {
			vector <Attribute*> tempAttrPtrs = HDF5Array (elements, datasetID, tempType, dimensions, dimSize);
			for (unsigned long int iAttrPtrs = 0; iAttrPtrs < tempAttrPtrs.size (); ++iAttrPtrs) {
				attrPtrs.push_back (tempAttrPtrs[iAttrPtrs]);
			}
		} else {
		if (tempDataclass == H5T_INTEGER) {
			vector <Attribute*> tempAttrPtrs = HDF5Integer (elements, datasetID, tempType, dimensions, dimSize);
			for (unsigned long int iAttrPtrs = 0; iAttrPtrs < tempAttrPtrs.size (); ++iAttrPtrs) {
				attrPtrs.push_back (tempAttrPtrs[iAttrPtrs]);
			}
		} else {
		if (tempDataclass == H5T_FLOAT) {
			vector <Attribute*> tempAttrPtrs = HDF5Float (elements, datasetID, tempType, dimensions, dimSize);
			for (unsigned long int iAttrPtrs = 0; iAttrPtrs < tempAttrPtrs.size (); ++iAttrPtrs) {
				attrPtrs.push_back (tempAttrPtrs[iAttrPtrs]);
			}
		} else {
		if (tempDataclass == H5T_STRING) {
			if (_myRank == 0) cerr << "*WARNING* " << memberName << ": Strings are not implemented in BASIN\n" << endl;
		} else {
		if (tempDataclass == H5T_COMPOUND) {
			vector <Attribute*> tempAttrPtrs = HDF5Compound (elements, datasetID, tempType, dimensions, dimSize);
			for (unsigned long int iAttrPtrs = 0; iAttrPtrs < tempAttrPtrs.size (); ++iAttrPtrs) {
				attrPtrs.push_back (tempAttrPtrs[iAttrPtrs]);
			}
		} else {
		if (tempDataclass == H5T_ENUM) {
			vector <Attribute*> tempAttrPtrs = HDF5Enum (elements, datasetID, tempType, dimensions, dimSize);
			for (unsigned long int iAttrPtrs = 0; iAttrPtrs < tempAttrPtrs.size (); ++iAttrPtrs) {
				attrPtrs.push_back (tempAttrPtrs[iAttrPtrs]);
			}
		} else {
		if (tempDataclass == H5T_REFERENCE) {
			if (_myRank == 0) cerr << "*WARNING* " << memberName << ": References are not implemented in BASIN" << endl;
		} else {
		if (tempDataclass == H5T_BITFIELD) {
			if (_myRank == 0) cerr << "*WARNING* " << memberName << ": Bitfields are not implemented in BASIN" << endl;
		} else {
		if (tempDataclass == H5T_OPAQUE) {
			if (_myRank == 0) cerr << "*WARNING* " << memberName << ": Opaques are not implemented in BASIN" << endl;
		} else {
		if (tempDataclass == H5T_VLEN) {
			if (_myRank == 0) cerr << "*WARNING* " << memberName << ": Variable Lengths are not implemented in BASIN" << endl;
		}}}}}}}}}}
		H5Tclose (tempType);
		H5Tclose (memberType);
	}
	H5Tclose (dataTypeID);
	return attrPtrs;
}


bool HDF5_IO::CreateFile () {
	hid_t plistID = H5Pcreate (H5P_FILE_ACCESS);
	H5Pset_fapl_mpiposix (plistID, BASIN_COMM, (hbool_t) false);
	_fileID = H5Fcreate(_fileName.c_str (), H5F_ACC_TRUNC, H5P_DEFAULT, plistID);
	H5Pclose (plistID);
	_groupID.push (H5Gopen (_fileID, "/"));
	_currentRegion.push (_data);
	return true;
}

bool HDF5_IO::CloseWrittenFile () {
	if (_currentRegion.empty ()) {
		if (_myRank == 0) cerr << "_currentRegion is empty when it should not be!" << endl;
		return false;
	}
	_currentRegion.pop ();
	if (!_currentRegion.empty ()) {
		if (_myRank == 0) cerr << "_currentRegion is not empty when it should be!" << endl;
		return false;
	}
	if (_groupID.empty ()) {
		if (_myRank == 0) cerr << "_groupID is empty when it should not be!" << endl;
		return false;
	}
	if (H5Gclose (_groupID.top ()) < 0) {
		if (_myRank == 0) cerr << "Error during H5Gclose ()!" << endl;
		return false;
	}
	_groupID.pop ();
	if (!_groupID.empty ()) {
		if (_myRank == 0) cerr << "_groupID is not empty when it should be!" << endl;
		return false;
	}
	if (H5Fclose (_fileID) < 0) {
		if (_myRank == 0) cerr << "Error during H5Fclose ()!" << endl;
		return false;
	}
	return true;
}

bool HDF5_IO::WriteDatasets () {
	map<string, Grid*>* grids = (*(_currentRegion.top ())).get_gridArray ();
	map<string, Grid*>::iterator gridIter;
	for (gridIter = (*grids).begin (); gridIter != (*grids).end (); ++gridIter) {
		WriteGrid(gridIter->second);
	}

	map<string, List*>* lists = (*(_currentRegion.top ())).get_listArray ();
	map<string, List*>::iterator listIter;
	for (listIter = (*lists).begin (); listIter != (*lists).end (); ++listIter) {
		WriteList(listIter->second);
	}
	return true;
}

bool HDF5_IO::WriteGrid (Grid* grid) {
	int numAttributes = (*grid).get_num_attributes ();
	for (unsigned int iAttr = 0; iAttr < numAttributes; ++iAttr) {
		Attribute* tempAttr = (*grid).get_attribute_pointer (iAttr);
		Attribute_Data_Type type = (*tempAttr).get_data_type ();

		if (type == BASIN_DATA_NONE) {
			return false;
		} else {
		if (type == BASIN_DATA_INT) {
			WriteIntegers (tempAttr);
		} else {
		if (type == BASIN_DATA_FLOAT) {
			WriteFloats (tempAttr);
		} else {
		if (type == BASIN_DATA_DOUBLE) {
			WriteDoubles (tempAttr);
		} else {
		if (type == BASIN_DATA_CHAR) {
			WriteCharacter (tempAttr);
		} else {
		if (type == BASIN_DATA_COMPLEX) {
			WriteComplex (tempAttr);
		} else {
		  if (type == BASIN_DATA_VECTOR2) {
#ifdef VECTOR23		    
			WriteVector2 (tempAttr);
#endif
		} else {
		if (type == BASIN_DATA_VECTOR3) {
#ifdef VECTOR23
			WriteVector3 (tempAttr);
#endif
		} else {
		}}}}}}}}
	}
	return true;
}

bool HDF5_IO::WriteList (List* list) {
	int numAttributes = (*list).get_num_attributes ();
	for (unsigned int iAttr = 0; iAttr < numAttributes; ++iAttr) {
		Attribute* tempAttr = (*list).get_attribute_pointer (iAttr);
		Attribute_Data_Type type = (*tempAttr).get_data_type ();

		if (type == BASIN_DATA_NONE) {
			return false;
		} else {
		if (type == BASIN_DATA_INT) {
			WriteIntegers (tempAttr);
		} else {
		if (type == BASIN_DATA_FLOAT) {
			WriteFloats (tempAttr);
		} else {
		if (type == BASIN_DATA_DOUBLE) {
			WriteDoubles (tempAttr);
		} else {
		if (type == BASIN_DATA_CHAR) {
			WriteCharacter (tempAttr);
		} else {
		if (type == BASIN_DATA_COMPLEX) {
			WriteComplex (tempAttr);
		} else {
		if (type == BASIN_DATA_VECTOR2) {
#ifdef VECTOR23
			WriteVector2 (tempAttr);
#endif
		} else {
		if (type == BASIN_DATA_VECTOR3) {
#ifdef VECTOR23
			WriteVector3 (tempAttr);
#endif
		} else {
		}}}}}}}}
	}
	return true;
}

bool HDF5_IO::WriteSubGroups () {
	unsigned int children = (*(_currentRegion.top ())).get_nchildren ();
	for (unsigned int iChild = 0; iChild < children; ++iChild) {
		ostringstream num;
		num << "#" << (iChild + 1);
		_currentRegion.push ((*(_currentRegion.top ())).get_child (iChild));
		_groupID.push (H5Gcreate (_groupID.top (), num.str ().c_str (), static_cast<size_t> (5)));

		WriteRegion ();

		H5Gclose (_groupID.top ());
		_groupID.pop ();
		_currentRegion.pop ();
	}
	return true;
}

bool HDF5_IO::WriteIntegers (Attribute* attr) {
	string name = (*attr).get_name ();
	vector<int> dims = (*attr).get_dims ();
	hsize_t* dimSize = new hsize_t[dims.size ()];
	for (unsigned int iDim = 0; iDim < dims.size (); ++iDim) {
		dimSize[iDim] = dims[iDim];
	}
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	hid_t dataspaceID = H5Screate_simple (dims.size (), dimSize, NULL);
	hid_t spaceIDMem;
	delete [] dimSize;

	hid_t datasetID = H5Dcreate (_groupID.top (), name.c_str (), H5T_NATIVE_INT, dataspaceID, H5P_DEFAULT);

	if (dims.size () > 1) {
		hsize_t* beginCoords = new hsize_t[dims.size ()];
		hsize_t* finalCoords = new hsize_t[dims.size ()];
		hsize_t* countCoords = new hsize_t[dims.size ()];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dims.size () - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dims[iDim]) % dims[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dims[iDim]) % dims[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			mod *= dims[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dims.size (), countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dims.size () == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}

	/* Retrieve data, for this process, from BASIN */
	int *data = new int[procElements];
	(*attr).get (procInds[0], procInds[1], data);

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);

	H5Dwrite(datasetID, H5T_NATIVE_INT, spaceIDMem, dataspaceID, plistID, data);

	delete [] data;
	H5Pclose (plistID);
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}

bool HDF5_IO::WriteFloats (Attribute* attr) {
	string name = (*attr).get_name ();
	vector<int> dims = (*attr).get_dims ();
	hsize_t* dimSize = new hsize_t[dims.size ()];
	for (unsigned int iDim = 0; iDim < dims.size (); ++iDim) {
		dimSize[iDim] = dims[iDim];
	}
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	hid_t dataspaceID = H5Screate_simple (dims.size (), dimSize, NULL);
	hid_t spaceIDMem;
	delete [] dimSize;

	hid_t datasetID = H5Dcreate (_groupID.top (), name.c_str (), H5T_NATIVE_FLOAT, dataspaceID, H5P_DEFAULT);

	if (dims.size () > 1) {
		hsize_t* beginCoords = new hsize_t[dims.size ()];
		hsize_t* finalCoords = new hsize_t[dims.size ()];
		hsize_t* countCoords = new hsize_t[dims.size ()];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dims.size () - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dims[iDim]) % dims[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dims[iDim]) % dims[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			mod *= dims[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dims.size (), countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dims.size () == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}

	/* Retrieve data, for this process, from BASIN */
	float *data = new float[procElements];
	(*attr).get (procInds[0], procInds[1], data);

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);
	H5Dwrite(datasetID, H5T_NATIVE_FLOAT, spaceIDMem, dataspaceID, H5P_DEFAULT, data);

	delete [] data;
	H5Pclose (plistID);
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}

bool HDF5_IO::WriteDoubles (Attribute* attr) {
	string name = (*attr).get_name ();
	vector<int> dims = (*attr).get_dims ();
	hsize_t* dimSize = new hsize_t[dims.size ()];
	for (unsigned int iDim = 0; iDim < dims.size (); ++iDim) {
		dimSize[iDim] = dims[iDim];
	}
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	hid_t dataspaceID = H5Screate_simple (dims.size (), dimSize, NULL);
	hid_t spaceIDMem;
	delete [] dimSize;

	hid_t datasetID = H5Dcreate (_groupID.top (), name.c_str (), H5T_NATIVE_DOUBLE, dataspaceID, H5P_DEFAULT);

	if (dims.size () > 1) {
		hsize_t* beginCoords = new hsize_t[dims.size ()];
		hsize_t* finalCoords = new hsize_t[dims.size ()];
		hsize_t* countCoords = new hsize_t[dims.size ()];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dims.size () - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dims[iDim]) % dims[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dims[iDim]) % dims[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			mod *= dims[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dims.size (), countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dims.size () == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}

	/* Retrieve data, for this process, from BASIN */
	double *data = new double[procElements];
	(*attr).get (procInds[0], procInds[1], data);

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);
	H5Dwrite(datasetID, H5T_NATIVE_DOUBLE, spaceIDMem, dataspaceID, H5P_DEFAULT, data);

	delete [] data;
	H5Pclose (plistID);
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}

bool HDF5_IO::WriteComplex (Attribute* attr) {
	string name = (*attr).get_name ();
	vector<int> dims = (*attr).get_dims ();
	hsize_t* dimSize = new hsize_t[dims.size ()];
	for (unsigned int iDim = 0; iDim < dims.size (); ++iDim) {
		dimSize[iDim] = dims[iDim];
	}
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	hid_t dataspaceID = H5Screate_simple (dims.size (), dimSize, NULL);
	hid_t spaceIDMem;
	delete [] dimSize;

	const hsize_t complexDim[1] = {2};
	hid_t arrayType = H5Tarray_create (H5T_NATIVE_DOUBLE, 1, complexDim, NULL);

	hid_t datasetID = H5Dcreate (_groupID.top (), name.c_str (), arrayType, dataspaceID, H5P_DEFAULT);

	/* Only create an HDF attribute with one process */
	if (_myRank == 0) {
		hid_t attributetypeID = H5Tcreate(H5T_INTEGER, static_cast<size_t> (sizeof (int)));
		hsize_t attributeDims[1] = {1};
		hid_t attributespaceID = H5Screate_simple(1, attributeDims, NULL);
		hid_t attributeID = H5Acreate(datasetID, "BASIN_COMPLEX", attributetypeID, attributespaceID, H5P_DEFAULT);

		int data[1] = {1};
		H5Awrite (attributeID, attributetypeID, data);

		H5Tclose (attributetypeID);
		H5Sclose (attributespaceID);
		H5Aclose (attributeID);
	}

	if (dims.size () > 1) {
		hsize_t* beginCoords = new hsize_t[dims.size ()];
		hsize_t* finalCoords = new hsize_t[dims.size ()];
		hsize_t* countCoords = new hsize_t[dims.size ()];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dims.size () - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dims[iDim]) % dims[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dims[iDim]) % dims[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			mod *= dims[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dims.size (), countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dims.size () == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}

	/* Retrieve data, for this process, from BASIN */
	fftw_complex *data = new fftw_complex[procElements];
	(*attr).get (procInds[0], procInds[1], data);

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);
	H5Dwrite(datasetID, arrayType, spaceIDMem, dataspaceID, H5P_DEFAULT, data);

	delete [] data;
	H5Pclose (plistID);
	H5Tclose (arrayType);
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}

bool HDF5_IO::WriteCharacter (Attribute* attr) {
	string name = (*attr).get_name ();
	vector<int> dims = (*attr).get_dims ();
	hsize_t* dimSize = new hsize_t[dims.size ()];
	for (unsigned int iDim = 0; iDim < dims.size (); ++iDim) {
		dimSize[iDim] = dims[iDim];
	}
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	hid_t dataspaceID = H5Screate_simple (dims.size (), dimSize, NULL);
	hid_t spaceIDMem;
	delete [] dimSize;

	hid_t datasetID = H5Dcreate (_groupID.top (), name.c_str (), H5T_NATIVE_CHAR, dataspaceID, H5P_DEFAULT);

	if (dims.size () > 1) {
		hsize_t* beginCoords = new hsize_t[dims.size ()];
		hsize_t* finalCoords = new hsize_t[dims.size ()];
		hsize_t* countCoords = new hsize_t[dims.size ()];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dims.size () - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dims[iDim]) % dims[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dims[iDim]) % dims[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			mod *= dims[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dims.size (), countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dims.size () == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}

	/* Retrieve data, for this process, from BASIN */
	char *data = new char[procElements];
	(*attr).get (procInds[0], procInds[1], data);

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);
	H5Dwrite(datasetID, H5T_NATIVE_CHAR, spaceIDMem, dataspaceID, H5P_DEFAULT, data);

	delete [] data;
	H5Pclose (plistID);
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}

#ifdef VECTOR23
bool HDF5_IO::WriteVector2 (Attribute* attr) {
	string name = (*attr).get_name ();
	vector<int> dims = (*attr).get_dims ();
	hsize_t* dimSize = new hsize_t[dims.size ()];
	for (unsigned int iDim = 0; iDim < dims.size (); ++iDim) {
		dimSize[iDim] = dims[iDim];
	}
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	hid_t dataspaceID = H5Screate_simple (dims.size (), dimSize, NULL);
	hid_t spaceIDMem;
	delete [] dimSize;

	const hsize_t complexDim[1] = {2};
	hid_t arrayType = H5Tarray_create (H5T_NATIVE_DOUBLE, 1, complexDim, NULL);

	hid_t datasetID = H5Dcreate (_groupID.top (), name.c_str (), arrayType, dataspaceID, H5P_DEFAULT);

	/* Only create an HDF attribute with one process */
	if (_myRank == 0) {
		hid_t attributetypeID = H5Tcreate(H5T_INTEGER, static_cast<size_t> (sizeof (int)));
		hsize_t attributeDims[1] = {1};
		hid_t attributespaceID = H5Screate_simple(1, attributeDims, NULL);
		hid_t attributeID = H5Acreate(datasetID, "BASIN_VECTOR2", attributetypeID, attributespaceID, H5P_DEFAULT);

		int data[1] = {1};
		H5Awrite (attributeID, attributetypeID, data);

		H5Tclose (attributetypeID);
		H5Sclose (attributespaceID);
		H5Aclose (attributeID);
	}

	if (dims.size () > 1) {
		hsize_t* beginCoords = new hsize_t[dims.size ()];
		hsize_t* finalCoords = new hsize_t[dims.size ()];
		hsize_t* countCoords = new hsize_t[dims.size ()];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dims.size () - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dims[iDim]) % dims[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dims[iDim]) % dims[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			mod *= dims[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dims.size (), countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dims.size () == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}

	/* Retrieve data, for this process, from BASIN */
	Vector2 *data = new Vector2[procElements];
	(*attr).get (procInds[0], procInds[1], data);

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);
	H5Dwrite(datasetID, arrayType, spaceIDMem, dataspaceID, H5P_DEFAULT, data);

	delete [] data;
	H5Pclose (plistID);
	H5Tclose (arrayType);
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}

bool HDF5_IO::WriteVector3 (Attribute* attr) {
	string name = (*attr).get_name ();
	vector<int> dims = (*attr).get_dims ();
	hsize_t* dimSize = new hsize_t[dims.size ()];
	for (unsigned int iDim = 0; iDim < dims.size (); ++iDim) {
		dimSize[iDim] = dims[iDim];
	}
	
	vector <int> procInds = (*attr).get_gen_distribution_index(_myRank);
	unsigned int procElements = procInds[1] - procInds[0] + 1;
	hid_t dataspaceID = H5Screate_simple (dims.size (), dimSize, NULL);
	hid_t spaceIDMem;
	delete [] dimSize;

	const hsize_t complexDim[1] = {2};
	hid_t arrayType = H5Tarray_create (H5T_NATIVE_DOUBLE, 1, complexDim, NULL);

	hid_t datasetID = H5Dcreate (_groupID.top (), name.c_str (), arrayType, dataspaceID, H5P_DEFAULT);

	/* Only create an HDF attribute with one process */
	if (_myRank == 0) {
		hid_t attributetypeID = H5Tcreate(H5T_INTEGER, static_cast<size_t> (sizeof (int)));
		hsize_t attributeDims[1] = {1};
		hid_t attributespaceID = H5Screate_simple(1, attributeDims, NULL);
		hid_t attributeID = H5Acreate(datasetID, "BASIN_VECTOR3", attributetypeID, attributespaceID, H5P_DEFAULT);

		int data[1] = {1};
		H5Awrite (attributeID, attributetypeID, data);

		H5Tclose (attributetypeID);
		H5Sclose (attributespaceID);
		H5Aclose (attributeID);
	}

	if (dims.size () > 1) {
		hsize_t* beginCoords = new hsize_t[dims.size ()];
		hsize_t* finalCoords = new hsize_t[dims.size ()];
		hsize_t* countCoords = new hsize_t[dims.size ()];

		/* Converts flattened array into n rank coordinates */
		int mod = 1;
		for (int iDim = dims.size () - 1; iDim >= 0; --iDim) {
			beginCoords[iDim] = ((procInds[0] / mod) + dims[iDim]) % dims[iDim];
			finalCoords[iDim] = ((procInds[1] / mod) + dims[iDim]) % dims[iDim];
			countCoords[iDim] = finalCoords[iDim] - beginCoords[iDim] + 1;
			mod *= dims[iDim];
		}

		/* Make hyperslab */
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* No longer needed, only for setting up hyperslab */
		delete [] beginCoords;
		delete [] finalCoords;
	
		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (dims.size (), countCoords, NULL);
		delete [] countCoords;
	} else {
	if (dims.size () == 1) {
		/* Make hyperslab */
		hsize_t beginCoords[1] = {procInds[0]};
		hsize_t countCoords[1] = {procElements};
		H5Sselect_hyperslab (dataspaceID, H5S_SELECT_SET, beginCoords, NULL, countCoords, NULL);

		/* Where the data from reading the hdf5 file will go */
		spaceIDMem = H5Screate_simple (1, countCoords, NULL);
	}}

	/* Retrieve data, for this process, from BASIN */
	Vector3 *data = new Vector3[procElements];
	(*attr).get (procInds[0], procInds[1], data);

	hid_t plistID = H5Pcreate (H5P_DATASET_XFER);
	H5Pset_dxpl_mpio (plistID, H5FD_MPIO_INDEPENDENT);
	H5Dwrite(datasetID, arrayType, spaceIDMem, dataspaceID, H5P_DEFAULT, data);

	delete [] data;
	H5Pclose (plistID);
	H5Tclose (arrayType);
	H5Sclose (dataspaceID);
	H5Dclose (datasetID);
	return true;
}
#endif
