#include "Region.h"
#include <vector>
#include <sstream>
#include "BASIN_IO.C"
using namespace std;


// The constructor for the region class.  In principle, since all regions 
// have a geometry, we will want to inherit the geometry/toplogy from the
// parent and set the bounds appropriately.
Region::Region(){
  myParent=NULL;
  myGeometry = Geometry();
  nChildren=0;
}

Region::Region(Region* parentRegion) {
  myGeometry=Geometry((*parentRegion).myGeometry);
  myParent=parentRegion;
  nChildren=0;
}

//Region::Region(string filename,int filetype) {
//  // 1 = GADGET
// if (filetype == 1) {
//    *(this)=read_gadget(filename);
//  }
//}

Region::Region(string filename, string fitsTemplate) {
  myGeometry=Geometry();
  nChildren=0;
  myParent=NULL;

  int myRank;
  MPI_Comm_rank(BASIN_COMM, &myRank );

  #ifdef BASIN_HAVE_HDF5
  if (filename.find (".h5") != string::npos) {
    HDF5_IO ioReader (filename, this);
    if (!ioReader.ReadFile ()) {
      if (myRank == 0) cout << "Error reading " << filename << "!" << endl;
    }
    return;
  }
  else if (filename.find(".bin") != string::npos) {
    BIN_IO ioReader (filename, this);
    if (!ioReader.ReadFile ()) {
      if (myRank == 0) cout << "Error reading " << filename << "!" << endl;
    }
    return;
  }
  #endif	

  int isFITS = 0;

#ifdef BASIN_HAVE_CFITSIO
  if(myRank == 0) {
    ifstream fin;
    fin.open(filename.c_str());
    string tempString;
    getline(fin, tempString);
    fin.close();
    if (tempString.find("SIMPLE")!=string::npos)
      isFITS = 1;
    else
      isFITS = 0;
  }
  MPI_Barrier(BASIN_COMM);
  
  MPI_Bcast(&isFITS, 1, MPI_INT, 0, BASIN_COMM);
  if(isFITS) {
    FITS_IO ioReader (filename, this);
    if(!ioReader.ReadFile ())
      if (myRank == 0) cout << "Error reading " << filename << "!" << endl;
  }
  
#endif
  if (!isFITS) {
	  TEXT_IO ioReader (filename, this);
	  if(!ioReader.ReadFile ())
	    if (myRank == 0) cout << "Error reading " << filename << "!" << endl;
  }
}

bool Region::SaveHDF5(string filename) {
#ifdef BASIN_HAVE_HDF5
	HDF5_IO ioOutput (filename, this);
	return ioOutput.WriteFile ();
#endif

	int myRank;
	MPI_Comm_rank(BASIN_COMM, &myRank );
	if (myRank == 0) cout << "HDF5 support was not compiled!" << endl;
	return false;
}

bool Region::SaveTEXT(string filename) {
  TEXT_IO ioOutput (filename, this);
  return ioOutput.WriteFile ();
}

//*********************************************************************
// Adding children, grids, etc.

// Add a child.  Since the basic structure of regions is assumed to be
// a tree, and since a DataSet, by definition, has only 1 root Region,
// the tree relationship is vital.  Therefore, adding a child to a
// node does the following:
// 1. Creates a new region (which inherits the geoemtry of the parent)
// 2. Makes that region a child of the parent
// 3. Makes the parent of the child the original region.
// 4. adds one to the number of children of the parent

// Add a child...

bool Region::add_child(Region* child) { 
  myChildren.push_back(child); 
  nChildren = myChildren.size (); 
  return true; 
}


Region* Region::add_child() {
  myChildren.push_back(new Region(this));
  nChildren++;
  return myChildren[nChildren-1];
}

Region* Region::add_child(string filename, string fitsTemplate) {
  myChildren.push_back(new Region(filename, fitsTemplate));
  nChildren++;
  return myChildren[nChildren-1];
}

// Add a List to a region
List* Region::add_list(string key,
		     int nElementsNew)	// default = 0
{
  listArray[key] = new List(this, nElementsNew);
  return listArray[key];
}

List* Region::add_list(string key, List* lNew)
{
  listArray[key] = lNew;
  lNew->set_parent(this);
  return listArray[key];
}

// Add a Grid to a Region


Grid* Region::add_grid(string key, const vector<int>& nxNew) {
  gridArray[key] = new Grid(this, nxNew);
  return gridArray[key];
}

Grid* Region::add_grid(string key, Grid* gNew) {
  gridArray[key]=gNew;
  gNew->set_parent(this);
  return gridArray[key];
}


//*********************************************************************
// Set the parameters for the grid...

// Define the geometry (including the boundaries)
int Region::set_geometry(int nDims, int metric, int boundary,int space,
			 const vector<double>& llNew, 
			 const vector<double>& urNew) {
  myGeometry = Geometry(nDims,metric,boundary,space,llNew,urNew);
  return 0;
}

// Another version to take strings rather than integers
int Region::set_geometry(int nDims, string metric, string boundary,string space,
			 const vector<double>& llNew, 
			 const vector<double>& urNew) {

  int metricType, boundaryType, spaceType;

  if(metric == "CARTESIAN") metricType = 0;
  else if(metric == "SPHERICAL") metricType = 1;
  else if(metric == "CYLINDRICAL") metricType = 2;
  else if(metric == "OTHER") metricType = -1;
  else metricType = 0;

  if(boundary == "ISOLATED") boundaryType = 0;
  else if(boundary == "PERIODIC") boundaryType = 1;
  else boundaryType = 0;

  if(space == "REAL") spaceType = 0;
  else if(space == "FOURIER_HERMITE") spaceType = 1;
  else if(space == "FOURIER_GENERAL") spaceType = 2;
  else spaceType = 0;

  //cerr << "metric: " << metricType << endl << "space: " << spaceType << endl << "boundary: " << boundaryType << endl;

  myGeometry = Geometry(nDims,metricType,boundaryType,spaceType,llNew,urNew);
  return 0;
}

//**********************************************************************

// getChild(int) return the pointer to the N-th child

Region* Region::get_child(int iChild) {
  if (iChild < nChildren && iChild >= 0) {
    return myChildren[iChild];
  }
  else {
    return NULL;
  }
}

// Is a particular point within a region?

int Region::calculate_inside(const vector<double>& x0) {
  int ret=1;
  vector<double> ll,ur;
  ll=myGeometry.get_bound(0);
  ur=myGeometry.get_bound(1);
  for (int i=0; i < myGeometry.get_ndims(); i++) {
    if (!((x0[i] > ll[i]) && (x0[i] < ur[i]))) {
      ret=0;
    }
  }
  return ret;
}

// Find the nearest particle...

Region*  Region::find_particle_nearest(string listKey,
				       const vector<double>& x0,
				       int* idx, int descendFlag) {
  Region* currRegion;

  if (descendFlag == 0) {
    *(idx)=get_list_pointer(listKey)->calculate_idx_nearest(x0);
    return this;
  }
  else {
    currRegion=find_in_region(x0);
    *(idx) = currRegion->get_list_pointer(listKey)->calculate_idx_nearest(x0);
    return currRegion;
  }
}

//Find nth nearest neighbor
Region* Region::find_particle_nth_nearest(string listKey,
					  const vector<double>& x0,
					  int* idx,
					  int particleDepth,
					  int descendFlag)
{
  Region* currRegion=NULL;
  if(descendFlag==0)
    {
      *idx=get_list_pointer(listKey)-> calculate_nth_nearest(particleDepth,x0);
      currRegion=this;
    }
  else
    {
      Region* currRegion= find_in_region(x0);
      *idx = currRegion->get_list_pointer(listKey)
			 ->calculate_nth_nearest(particleDepth,x0);
    }
  return currRegion;
}

Region* Region::find_in_region(const vector<double>& x0) {
  int i,isDone=0,changed; 
  Region *currRegion;

  // Scroll through the various regions, and find the lowest level
  // one which contains the point.

  currRegion=this;
  i = this->calculate_inside(x0);
  if (i == -1) {
    return NULL;
  }
  else {
    while (isDone == 0) {
      changed=0;
      for (int j=0; j < currRegion->get_nchildren() ; j++) {
	if (currRegion->get_child(j)->calculate_inside(x0) ) {
	  currRegion = currRegion->get_child(j);
	  changed=1;
	  break;
	}
      }  
      if ((!changed)||(currRegion->get_nchildren() == 0)) {
	isDone=1;
      }
    }
    return currRegion;
  }
}

// Display the contents of a region.
// If resolution < 0 (default), recursively descend to display all child
// regions; otherwise, resolution sets the number of levels to descend.

// Show the history of the Region


// Delete a Grid from a Region

int Region::delete_grid(string key)
{ 
  gridArray.erase(key);
  return 0;
}

//Delete a List from a Region

int Region::delete_list(string key)
{
  listArray.erase(key);
  return 0;
}

bool Region::change_grid_key (string oldkey, string newkey) { 
	map <string, Grid*>::iterator iter = gridArray.find (oldkey); 
	if (iter != gridArray.end ()) { 
		gridArray.insert (make_pair (newkey, iter->second)); 
	} else { 
		return false; 
	} 
	gridArray.erase (iter->first); 
	return true; 
} 

bool Region::change_list_key (string oldkey, string newkey) { 
	map <string, List*>::iterator iter = listArray.find (oldkey); 
	if (iter != listArray.end ()) { 
		listArray.insert (make_pair (newkey, iter->second)); 
	} else { 
		return false; 
	} 
	listArray.erase (iter->first); 
	return true; 
} 

bool Region::delete_child (Region* reg) { 
	vector <Region*>::iterator iter; 
	for (iter = myChildren.begin (); iter != myChildren.end (); ++iter) { 
		if ((*iter) == reg) { 
			myChildren.erase (iter); 
			nChildren = myChildren.size (); 
			return true; 
		}
	}
	return false; 
}

Region* Region::get_child (string childName) { 
	for (unsigned int i = 0; i < nChildren; ++i) { 
		if ((*myChildren[i]).get_name () == childName) { 
			return myChildren[i]; 
		} 
	} 
	return NULL; 
}

string Region::get_client_information (string info, string depth) { 
	info += depth + "REGION:" + name + "\n"; 

	map <string, List*>::iterator iterList; 
	for (iterList = listArray.begin (); iterList != listArray.end (); ++iterList) { 
		info += depth + "^LIST:" + iterList->first + "\n"; 
		info = (*iterList->second).get_client_information (info, depth + "^"); 
	} 

	map <string, Grid*>::iterator iterGrid; 
	for (iterGrid = gridArray.begin (); iterGrid != gridArray.end (); ++iterGrid) { 
		info += depth + "^GRID:" + iterGrid->first + "\n"; 
		info = (*iterGrid->second).get_client_information (info, depth + "^"); 
	} 

	for (unsigned int i = 0; i < nChildren; ++i) { 
		info = (*myChildren[i]).get_client_information (info, depth + "^"); 
	} 

	return info; 
}

/*Region::~Region(){
  if( myParent != NULL );
	  //		myParent->remove_child( this );

	map< string, List* >::iterator listit;
	for( listit = listArray.begin(); listit!= listArray.end(); listit++ ) {
		delete listit->second;
	}

        map< string, Grid* >::iterator gridit;
        for( gridit = gridArray.begin(); gridit!= gridArray.end();gridit++ ) {
                delete gridit->second;
        }

	for( int i=0;i < myChildren.size(); i++ ) {
		myChildren[i]->myParent = NULL;
		delete myChildren[i];
	}

}
*/
