/**
 * This file is a demonstration of how the VisIt module package can be used
 * in C++ code.
 */

#include "basin.h"
#include "visit_module.h"
#include "cosmology.h"
#include "transforms.h"
#include "statistics.h"
#include "Attribute.h"

int main(int argc, char **argv) {
    
    basin_initialize();

    Attribute mx, my, mz, mr;

    // Generate some random data to display
    mx = attribute_random(2000, 1);
    my = mx;
    mz = attribute_random(2000, 2);
    mr = attribute_random(2000, 3);

    // Now start the visit_module package.
    // Behind the scenes, this is creating a new thread
    // that will listen for communications from VisIt
    visit_module_start();

    // Make a variable named 'var0' on a mesh generated from mx, my, and mz
    // and expose it to visit
    visit_expose_discrete(mx, my, mz, mr, "var0");

    // Now make some bogus data for a 2d plot
    Attribute mx2, my2, mr2;
    mx2 = attribute_random(1000, 1);
    my2 = mx2;
    mr2 = mx2;

    // Make a variable named 'var1' on a mesh generated from mx2, my2 and
    // expose it to visit
    visit_expose_discrete(mx2, my2, mr2, "var1");

    // Now let's build a 3d grid
    vector<int> nx(3, 10);

    Attribute a(nx, double(0.0));
    vector<int> pos(3, 0);
    for(pos[0] = 0; pos[0] < 10; pos[0]++) {
        for(pos[1] = 0; pos[1] < 10; pos[1]++) {
            for(pos[2] = 0; pos[2] < 10; pos[2]++) {
                double d = 10*10*pos[0] + 10*pos[1] + pos[2];
                a.set_element(pos, d);
            }
        }
    }

    // Define the dimensions of our grid
    vector<double> ll(3, 0.0);
    vector<double> ur(3, 1.0);

    // Expose this as a rectilinear mesh built using the properties
    // of the grid. 'a' contains the data for the mesh
    visit_expose_continuous(ll, ur, nx, a, "var2");

    // We can also use the Grid class to make this easier
    Grid g(nx);
    g.set_context(ll, ur, nx);
    g.add_attribute("a", a);

    // Now the mesh is built using the grid's context information
    visit_expose_continuous(g, a, "var3");

    
    
    // Now let's try exposing data from some more realistic calculations
    Region catalog("zcatalog.dat");
    List *catlist = catalog.get_list_pointer();

    equatorial_to_xyz(catlist, "ra", "dec", "z");

    // Show the object locations on a point mesh
    // positions are generated from the coords in x_conv, y_conv, and z_conv
    // our data is the "z" value of the position. "Locations" is the name
    // of the data that VisIt will use.
    visit_expose_discrete(*catlist, "x_conv", "y_conv", "z_conv", "z", "Locations");

    
    // Calculate the density of the field 
    vector<int> nx2(3, 50);
    Grid *dgrid = compute_density(catlist, "x_conv", "y_conv", "z_conv", "dens_real", nx2, 3);

    // Send to visit the result of the density calculation
    // dgrid defines the mesh, "dens_real" is the name of the grid's
    // member attribute that we will extract data from and "Density Real"
    // is the name that visit will use for this data
    visit_expose_continuous(*dgrid, "dens_real", "Density Real");


    // Tell our demo not to exit while VisIt is connected
    visit_module_join();
    
    // Clean up before process exits
    basin_finalize();
}

