#include "basin_package.h"
#include "cosmology.h"

cosmology::cosmology() {
  omega_m=0.27;
  omega_l=0.73;
  omega_b=0.02;
  omega_r=1.e-4;
  T_0=2.73;
  h=0.71;
  n_s=0.95;
  sig_8=1.0;  
  w=-1;
  norm=1;
  normset=0;
  k_curv=0;
}

Attribute cosmology::hubble(Attribute& a) {
  // Compute the Hubble constant at a particular expansion factor
  // Always expressed in terms of h.

  // We're doing this at near the present epoch. 
  double omega_k=1.-omega_m-omega_l;

  return(100.*sqrt(omega_m/pow(a,3)+omega_l+omega_k/pow(a,2)));
}

double cosmology::hubble(double a) {
  double omega_k=1.-omega_m-omega_l;
  return(100.*sqrt(omega_m/pow(a,3)+omega_l+omega_k/pow(a,2)));
}


Attribute* cosmology::redshift_to_growth(List* listptr,string zname,
                                         string dname) {
  Attribute* z_attr;
  if (listptr->attribute_exist(zname)) {
    z_attr=listptr->get_attribute_pointer(zname);
  } else {
    z_attr=listptr->get_attribute_pointer(0);
  }
  Attribute d_attr=redshift_to_growth(*z_attr);
  listptr->add_attribute(dname,d_attr);
  MPI_Barrier(BASIN_COMM);
  return listptr->get_attribute_pointer(dname);
}

Attribute cosmology::redshift_to_growth(Attribute &z) {
  // Computes the linear growth parameter at redshift z.
  int nbins=10000;
  double da=1./nbins;
  double a;
  double d[nbins+1];
  double tot=0;
  d[0]=0.;
  double h1;
  for (int i=1; i <= nbins; i++) {
    a=da*i;
    h1=hubble(a)/100.;
    tot+=da/pow(a*h1,3);
    d[i]=2.5*omega_m*h1*tot;
  }

  // Finish it up by getting the local distribution...
  int myid;
  MPI_Comm_rank(BASIN_COMM, &myid);
  int hi,lo;
  z.get_gen_distribution(myid,&lo,&hi);
  double z_loc[hi-lo+1];
  z.get(lo,hi,z_loc);
  double d_loc[hi-lo+1];
  double a1,w;
  int ibin;
  for (int i=0; i < hi-lo+1; i++) {
    a=1./(1+double(z_loc[i]));

    ibin=int(a/da);
    w=1.-(a-ibin*da);
    if ((ibin < nbins)&&(ibin >= 0)) {
      d_loc[i]=w*d[ibin]+(1.-w)*d[ibin+1];
    } else if (ibin < 0) {
      d_loc[i]=d[0];
    } else {
      d_loc[i]=d[nbins];
    }
  }
  Attribute d_attr(z);
  MPI_Barrier(BASIN_COMM);
  d_attr.put(lo,hi,d_loc);
  MPI_Barrier(BASIN_COMM);
  return d_attr;
}

double cosmology::redshift_to_growth(double z) {

  // Computes the linear growth parameter at redshift z.
  int nbins=10000;
  double amax=1/(1.+z);
  double da=amax/nbins;
  double a;
  double d=0;
  double tot=0;
  double h1;
  for (int i=1; i <= nbins; i++) {
    a=da*i;
    h1=hubble(a)/100.;
    tot+=da/pow(a*h1,3);
    d=2.5*omega_m*h1*tot;
  }
  return d;
}


Attribute* cosmology::redshift_to_distance(List* listptr,string zname,string rname,
					   int distance_type,int hflag) {
  Attribute* z_attr;
  // Does the name already exist?
  if (listptr->attribute_exist(zname)) {
    z_attr=listptr->get_attribute_pointer(zname);
  } else {
    z_attr=listptr->get_attribute_pointer(0);
  }
  Attribute r_attr=redshift_to_distance(*z_attr,distance_type,hflag);
  listptr->add_attribute(rname,r_attr);
  return listptr->get_attribute_pointer(rname);
}


Attribute& cosmology::redshift_to_distance(Attribute& z,int distance_type, int hflag){
  // Measures the various distances from the observer to a particular redshift

  // distance_type 
  // ------------
  // 0 : Comoving Distance
  // 1 : Proper Distances
  // 2 : Angular diameter distance
  // 3 : Luminosity Distance

  // hflag
  // 0 : express the result in h^-1 Mpc
  // 1 : express the result in h Mpc

  double hmult;
  if (hflag == 0) {
    hmult=1./h;
  } else {
    hmult=1;
  }

  double c=3.e5;
  double zmin=min(z);
  double zmax=max(z);
  double amax=1.1/(1.+zmin);
  double amin=0.5/(1.+zmax);
  int nsteps=1000;
  vector<double> chi(nsteps,double(0.));
  double da=(amax-amin)/(nsteps-1);
  double a=amin;
  // We need to calculate chi first in any case.
  chi[0]=c*da*(1./(pow(amin,2)*hubble(amin)+0.01))/2.;
  for (int istep=1; istep < nsteps-1; istep++) {
    a=a+da/2.;
    chi[istep]=chi[istep-1]+c*da/(a*a*hubble(a));
    a=a+da/2.;
  }
  int id0;
  double w;
  Attribute *r=new Attribute(z.size(),double(0.));
  int lo,hi,basin_id;
  MPI_Comm_rank(MPI_COMM_WORLD,&basin_id);
  r->get_gen_distribution_index(basin_id,&lo,&hi);
  int type=z.get_data_type();
  double* zdat;
  if (type == BASIN_DATA_DOUBLE) {
    zdat=(double*) malloc(sizeof(double)*(hi-lo+1));
    z.get(lo,hi,zdat);
  } 
  else if (type == BASIN_DATA_FLOAT) {
    float *zdat1=(float*)malloc(sizeof(float)*(hi-lo+1));
    zdat=(double*) malloc(sizeof(double)*(hi-lo+1));
    z.get(lo,hi,zdat1);
    for (int i=0; i < hi-lo+1;i++) {
      zdat[i]=zdat1[i];
    }
    free(zdat1);
  }
  double *rdat=(double*) malloc(sizeof(double)*(hi-lo+1));
  // What is chi0?
  id0=int((1.-amin)/da);
  w=(1.-amin)/da-id0;
  double chi0=chi[id0]*(1.-w)+w*chi[id0+1];
  for (int i=0; i < hi-lo+1; i++) {
    id0=int((1./(1.+zdat[i])-amin)/da);
    w=(1./(1.+zdat[i])-amin)/da-id0;
    rdat[i]=chi0-(chi[id0]*(1.-w)+w*chi[id0+1]);
    if (distance_type == 0) {
      rdat[i]=rdat[i]*hmult;
    } else if (distance_type == 1) {
      rdat[i]=rdat[i]*hmult/(1.+zdat[i]);
    } else {
      if (k_curv==0) {
	// Flat universe.
	if (distance_type == 2) {
	  rdat[i]=rdat[i]*hmult/(1.+zdat[i]);
	} else {
	  rdat[i]=rdat[i]*hmult*(1.+zdat[i]);
	}
      } else {
	// nonflat universe 
	double R0=c/(100.*abs(1.-omega_m-omega_l));
	if ((k_curv == 1)&&(distance_type == 2)) {
	  rdat[i]=hmult*R0*sin(rdat[i]/R0)/(1.+zdat[i]);
	} else if ((k_curv==1)&&(distance_type ==3)){
	  rdat[i]=hmult*R0*(1.+zdat[i])*sin(rdat[i]/R0);
	} else if ((k_curv == -1)&&(distance_type == 2)) {
	  rdat[i]=hmult*R0*sinh(rdat[i]/R0)/(1.+zdat[i]);
	} else if ((k_curv == -1)&&(distance_type ==3)){
	  rdat[i]=hmult*R0*(1.+zdat[i])*sinh(rdat[i]/R0);
	}
      }
      
    }
  }
  r->put(lo,hi,rdat);
  free(rdat);
  free(zdat);
  MPI_Barrier(BASIN_COMM);
  return *r;
}


int cosmology::normalize() {
  if (normset == 1) {
    return -1;
  }
  normset=1;
  double L=200.; // h^-1 Mpc, kmin=2*Pi/L
  double R=8.;
  double kmin=2.*3.14159/L;
  double kmax=2.*3.14159/0.1;
  int nsteps=200;
  double dk=(kmax-kmin)/(nsteps-1);
  double tot=0;
  for (int i=0; i < nsteps; i++) {
    double k=kmin+i*dk;
    double y=R*k;
    double fk=3*(sin(y)-y*cos(y))/pow(y,3);
    tot+=power_spectrum(k,0)*k*k*dk*fk*fk;
  }
  tot=tot/(2.*3.14159*3.14159);
  cerr << tot << "\n";
  // Now, tot should be sig8^2, so:
  norm=sig_8*sig_8*norm/tot;
  return 0;
}

double cosmology::power_spectrum(double k, double z) {
  if (normset == 0) normalize();
  double q=k*pow(T_0/2.7,2)/
    (omega_m*h*h*exp(-omega_b-sqrt(h/0.5)*omega_b/omega_m));
  double T=(log(1+2.34*q)/(2.34*q))*pow(1+3.89*q+pow(16.1*q,2)+pow(5.4*q,3)+
				     pow(6.71*q,4),-0.25);

  
  double D=redshift_to_growth(z)/redshift_to_growth(0.);
  double P=D*D*norm*pow(k,n_s)*T*T;
  // Of course, this is not redshift corrected.  Correct with the growth 
  // factor.
  return(P);
}

vector<double> cosmology::power_spectrum(vector<double> k, double z) {
  int npts=k.size();
  vector<double> pk(npts);
  for (int i=0; i < npts; i++) {
    pk[i]=power_spectrum(k[i],z);
  }
  return pk;
}




Attribute& cosmology::power_spectrum(Attribute& k,double z){
  Attribute* P=new Attribute(k);
  if (normset == 0) normalize();

  //Dear Enrico, Once again, I make the explicit assumption that k is
  //a type double.  If not, that's too bad.
  Attribute q=k*pow(T_0/2.7,2)/
    (omega_m*h*h*exp(-omega_b-sqrt(h/0.5)*omega_b/omega_m));
  Attribute T=(log(1+2.34*q)/(2.34*q))*pow(1+3.89*q+pow(16.1*q,2)+pow(5.4*q,3)+
				     pow(6.71*q,4),-0.25);
  double D=redshift_to_growth(z)/redshift_to_growth(0.);
  (*P)=D*D*norm*pow(k,n_s)*T*T;

  return *(P);
}
