/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*! \file Region.h
    \brief Contains the Region class.

    Region is the upper-most object that organizes the dataset into a tree hierarchy.
*/


#ifndef REGION_H
#define REGION_H

#include "Geometry.h"
#include "Data.h"

class List;
class Grid;
using namespace std;

/** \class Region Region.h "Region.h"
    \brief Region is the top-most object, containing \link List Lists\endlink and \link Grid Grids\endlink.
    
    A Region may have associated with it a number of different \link List Lists\endlink and \link Grid Grids\endlink.  
    It also contains a Geometry.
*/
class Region{
 public:
  /** @name Region Constructors
      Functions to construct Region objects. All nodes must call a constructor to work with a Region.*/
  //@{
  //!Empty Constructor
  Region();
  /**< Creates a blank Region with no parent or children and a default Geometry (0 dimensional, CARTESIAN, ISOLATED, REAL space).*/


  // Region constructor which sets the parent Region and Geometry
  Region(Region* parentRegion);
  /**< Creates a Region with no children, setting myParent to parentRegion and copying the parent's Geometry.  \warning This constructor should probably not be called directly, since no action is taken to link the new Region into the parent's myChildren.  Instead, use Region::add_child().*/

  //!Region constructor for reading in a file of specific (simulation specific) format
  //Region(string filename, int filetype);

  //!Region constructor for reading in a file and constructing
  Region(string filename, string fitsTemplate="");
  //!
  Region(string filename, int firstSnap, int lastSnap, int snapStep, string fitsTemplate="");

  bool SaveHDF5(string filename);
  bool SaveTEXT(string filename);
  /**< This function will attempt to determine the appropriate read function to use and construct the Region based on the data found.  If it fails to find the HDF hierarchy file or a proper FITS file, it will attempt to Region::read_txt as a last resort.*/

  //@}
  /** @name Region Input Functions
      Functions that do the actual reading of the file of a specific type.*/
  //@{
  // HDF files, using Enzo Simulation conventions
  /**< \bug This function is currently not very robust.  Believe it may ONLY work for Enzo formatted data.*/

  // FITS files, using SDSS conventions
  /**< This function will read in a FITS file, using a template file to set the position attributes and determine which parts of the file are wanted or unwanted. \bug The template feature is currently disabled.  So this will currently have issues with data not containing "ra" and "dec" members.*/

  /**< The header conventions are as follows:\n
     [Goldberg or Ryan inserts information here]\n
  */

  //@}

  /** @name Region Tree Hierarchy Functions
      Functions that deal with creating and traversing the tree hiearchy.*/
  //@{
  //! Add an empty child to the Region
  Region* add_child();

  bool add_child(Region*);

  bool delete_child (Region*);
  /**< This function creates a new Region with Region::Region(Region* parentRegion), and places the new pointer in myChildren.*/
  //! Create a child form a file then attach it to the Region
  Region* add_child(string filename, string fitsTemplate="");
  /**< This function creates a new Region with Region::Region(string filename, string fitsTemplate=""), and places the new pointer in myChildren. \warning It is possible for the file input function to overwrite the Geometry and thus make it different from the parent's.  However, if it is at all logical to attach the two Region's through this relationship, that shouldn't happen.*/

  //! Return the pointer to the parent Region
  Region* get_parent() {return myParent;};

  //! Return the pointer to a particular child
  Region* get_child(int iChild);
  Region* get_child(string);
  /**< Accessor function for a specific element of myChildren.  Returns NULL pointer if iChild is not in bounds.*/

  //! How many children are there?
  int get_nchildren() {return nChildren;};
  //@}

  /** @name Basic List and Grid Manipulation
   Primarily, Regions contain Lists and Grids.  Here are some
    basic functions to create and scroll though them */
  //@{
  // Add a List to a Region
  List* add_list(string key, int nElementsNew=0);
  /**< Adds a List to listArray with index=key.  The constructor List::List(Region* parent, int nElementsNew) is used to create it.*/

  //! Using an existing pointer, adds a List to a Region
  List* add_list(string key, List* lNew);
  /**< Adds List specified by lNew to the Region with index=key.*/

  /// Add a Grid to a Region
  Grid* add_grid(string key, const vector<int>& nxNew);
  /**< Adds a Grid to gridArray with index=key.  The constructor Grid::Grid(Region* parent, const vector<int>& nxNew) is used to create it.*/

  //! Using an existing pointer, adds a Grid to a Region
  Grid* add_grid(string key, Grid* gNew);
  /**< Adds Grid specified by gNew to the Region with index=key.*/

  //! return a pointer to an attribute
  Attribute* get_attribute_pointer(string key, string listkey="") {
    List* mylist=get_list_pointer(listkey);
    return (mylist -> get_attribute_pointer(key));
  }



  //! return a copy of an attribute
  Attribute copy_attribute(string key, string listkey="") {
    List* mylist=get_list_pointer(listkey);
    return (mylist -> copy_attribute(key));
  }



  //! Return a pointer to a particular list
  List* get_list_pointer(string key="RegionList0") {
    if(listArray.find(key) != listArray.end())
      return listArray[key];
    else {
      if(listArray.size()!=0)
	return listArray[listArray.begin()->first];
      else{
	cerr << "Could not find List " << key << endl;
	return NULL;
      }
    }
  };

  int query_list_key(string key) {
    if (listArray.find(key) != listArray.end()) {
      return 1;
    } else {
      return 0;
    }
  };

  //! Return a pointer to a particular grid
  Grid* get_grid_pointer(string key="RegionGrid0") {
    if(gridArray.find(key) != gridArray.end())
      return gridArray[key];
    else {
      if(gridArray.size()!=0)
	return gridArray[gridArray.begin()->first];
      else{
	cerr << "Could not find Grid " << key << endl;
	return NULL;
      }
    }
  }

  //! How many Lists are there?
  int get_nlists() {return listArray.size();};

  //! How many Grids are there?
  int get_ngrids() {return gridArray.size();};

  //! Delete a list
  int delete_list(string key);

  //! Delete a grid
  int delete_grid(string key);

  bool change_grid_key (string oldkey, string newkey);

  bool change_list_key (string oldkey, string newkey);
  //@}
  /** @name Geometry and Boundary Functions
      These functions query the associated Geometry object through the Region.*/
  //@{
  /// Define the geometry (including the boundaries)
  int set_geometry(int nDims, int metric, int boundary, int space, const vector<double>& llNew, const vector<double>& urNew);
  int set_geometry(int nDims, string metric, string boundary, string space, const vector<double>& llNew, const vector<double>& urNew);

  // Return either the lower left(0) or upper right(1) boundary 
  vector<double> get_bound(int iSide) {return myGeometry.get_bound(iSide);};
 
  /// Return a pointer to the geometry
  Geometry* get_geometry() {return &myGeometry;};

  //@}

  /** @name Region History Functions
      Functions to add or examine the history of a Region.*/
  //@{
  
  /// Add Event to list
  /** Any function that alters a Region should call this function to indicate which function was called, with what parameters, and what the exit status was.  For example:
\code
int some_func(Region* a, ...){
   a->add_event("some_func called with parameters...");
   //function code
   if(error_flag){
      a->add_event("some_func exited with -1 due to...");
      return -1;
   }
   //more code
   a->add_event("some_func completed successfully");
   return 0;
}
\endcode
*/




 
  /// Return pointer to Event list


  // Display the contents of the myHistory list
  Region* display_history_contents(ostream& s=cout);
 
  //@}
  /** @name Basic Calculation Functions
      Functions to perform basic data calculations.*/
  //@{
  /// Is a position inside the region or not?
  int calculate_inside(const vector<double>& x0);


  /// Find the nearest particle to a particular coordinate.   
  Region* find_particle_nearest(string listKey, const vector<double>& x0, int* idx, int descendFlag);
  /**< Take as arguments the List name, a pointer into which to put
   *  the final index, and a flag=0 indicating don't descend, or
   *  flag=1, indicating descend into lower levels.  Return the
   *  pointer to the Region where the particle is.*/

  /// Find the nth nearest neighbor to a particle.
  /** You must specify the List name (string key), and the index of the
   * particle that we start at, the int n representing the nth nearest
   * neighbor, and a 0 or 1 int flag indicates descent to lower levels
   * 1=descend, 0=stay in current region.
   */
  Region* find_particle_nth_nearest(string listKey,const vector<double>& x0,int* idx,int particleDepth,int descendFlag);

  /// Find the lowest-level region in a tree containing a
  /// particular coordinate.
  Region* find_in_region(const vector<double>& x0);
  //@}
  /** @name Display Functions
      Functions to display aspects of the Region.*/
  //@{
  // Display the contents of the Region (and its children)
  int display_region_contents(data_type myType=NONE,
				  int resolution=-1,
				  ostream& s=cout,
				  string searchString="");
  /**< data_type specifies what type of data you want to filter for.  resolution sets the number of levels to iterate down.  resolution=-1 will display the entire tree.  s sets where the display will occur.  Lastly, searchString will filter to just display a specific attribute in each Region.*/

  // Represents the object.
  string represent(){
    stringstream ss;
    display_region_contents(NONE, -1, ss);
    return ss.str();
  }
  /**< Currently only displays the contents, but can be expanded to include other useful information.  This becomes the Python __repr__ function. */

  //@}
  /** @name Density Computation Functions
      These functions compute the density of an attribute in a Region's List.*/
  //@{
  // Computes the density of attribute listKey in List listName and returns a Grid with attribute gridKey. 
  //Grid* descend_region_density(string gridKey="", string listKey="", string listName="RegionList",
  //			       vector<int> nxNew=BASIN_nxDefault, int iSmooth=0,
  //			       vector<float> ll=BASIN_posDefault, vector<float> ur=BASIN_posDefault);
  /*!< To be run on all regions.  Take a given region, call the 
   compute_density function, and iteratively call the function on
   all of its children.  Then combine the data from the children.*/

  /// Computes the density of the entire Region and all children, returning a Graphics_Grid.
  //Graphics_Grid compute_density_combined(string gridKey, string listKey, string gridName, string listName,vector<int> nxNew, int iSmooth);
  /**< This runs #descend_region_density, attaches the returned Grid as gridName, and returns a Graphics_Grid of it.  The boundaries are obtained from the Geometry attached to the Region and must be properly set.*/

  //Note: Currently implementation is missing..?
  //@}

  //Returns the maps of Grids and Lists to the user... dangerous?
  map<string,Grid*>* get_gridArray() { return &gridArray; }
  map<string,List*>* get_listArray() { return &listArray; }

  string get_gridNames () {
    string names = "";
    map<string, Grid*>::iterator iter;
    for (iter = gridArray.begin (); iter != gridArray.end (); iter++) {
    	names += iter->first + "\n";
    }

    return names;
  }

  string get_listNames () {
    string names = "";
    map<string, List*>::iterator iter;
    for (iter = listArray.begin (); iter != listArray.end (); iter++) {
    	names += iter->first + "\n";
    }

    return names;
  }

  void set_name (string Name) {
  	name = Name;
  }

  string get_name () {
  	return name;
  }

  string get_client_information (string info = "", string depth = "");

 private:

  string name;
  /// Associated Geometry object
  /** myGeometry defines the Geometry of the Region.  Note that this Geometry is not always the same as that of a contained List or Grid.*/
  Geometry myGeometry; 
  /// list of \link Event Events\endlink that have been logged
  /**myHistory is written to whenever a function does something on the Region.  This log can be used to reproduce the current state of the dataset from the beginning.*/
  /// Map of contained \link List Lists\endlink
  /** Note that the Region actually contains the List, not merely a pointer to it.*/
  map < string, List* > listArray;
  ///Map of contained \link Grid Grids\endlink
  /** Note that the Region actually contains the Grid, not merely a pointer to it.*/
  map < string, Grid* > gridArray;
  /// Upper level Region that should contain this one as a child
  /** The Region inherits the Geometry of its parent. Descending calls made on this parent Region will affect the child as well.*/
  Region *myParent; 
  /// Vector containing pointers to the  child Regions
  /** These are the Regions spawned from this Region, sharing Geometry and participating in any descending calls made on this Region.*/
  vector<Region*> myChildren;
  /// Current count of child Regions
  /** Equivalent to int(myChildren.size())*/
  int nChildren;

};

#endif
