/*
 *  Simple C++ gnuplot API for use with StarCluster tools.
 *  Copyright (C) 2003  StarCluster team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#ifdef USE_ERRC
	#include <err.h>
#else
	#include <error.h>
#endif
#include <cerrno>
#include <float.h>
#include <unistd.h>
#include <cstdio>
#include <fstream>
#include <cstdlib>
#include "gnuplot.h"

#define INVALID_TIME FLT_MIN

gnuplot::gnuplot(const bool persist) {
  if (!(fp = fp_ =
	popen(persist ? "gnuplot -noraise -persist" : "gnuplot -noraise", "w")))
#ifdef USE_ERRC
	errc(EXIT_FAILURE, errno, "can't create pipe/process: popen(3)");
#else
	error(EXIT_FAILURE, errno, "can't create pipe/process: popen(3)");
#endif
  autoscale = data_only = multiplot = plot3d = false;
  first_dataset = first_flush = first_number = new_plot = true;
  graphs = plots[0] = plots[1] = 1;
  graph = n = pause = plot = 0;
  style = "points";
  time = INVALID_TIME;
  title = xlabel = ylabel = zlabel = "";
}

gnuplot::~gnuplot() {
  if (fp == fp_ && pclose(fp) == -1)
#ifdef USE_ERRC
      	errc(EXIT_FAILURE, errno, "can't close pipe/process: pclose(3)");
#else
     	error(EXIT_FAILURE, errno, "can't close pipe/process: pclose(3)");
#endif
}

gnuplot& gnuplot::flush(const bool to_pause) {
  if (new_plot && !data_only) fflush(fp);
  else {
    fputs(data_only ? "\n" : "e\n", fp), fflush(fp);
    for (size_t i = 0; i < buf.size(); ++i) {
      for (size_t j = 0; j < buf[i].size(); j += (plot3d?3:2))
	if (plot3d)
	  fprintf(fp, "%g %g %g\n", buf[i][j], buf[i][j+1], buf[i][j+2]);
	else fprintf(fp, "%g %g\n", buf[i][j], buf[i][j+1]);
      fputs(data_only ? "\n" : "e\n", fp), fflush(fp);
      buf[i].clear();
    }
  }
  if (!data_only) {
    if (first_flush && !autoscale && !new_plot) {
      fputs("set xrange restore\nset yrange restore\nset zrange restore\n", fp);
      first_flush = false;
    }
    if (!first_buf.empty()) {
      if (multiplot) {
	const float relsize = plot3d ? 1 : 0.9;
	fprintf(fp, "set size 1,1\nset origin 0,0\nclear\nset size %g,%g\n",
		relsize/plots[0], relsize/plots[1]);
	plot = 0;
      }
      new_plot = true;
      first_dataset = false;
      for (size_t i = 0; i < first_buf.size(); *this << first_buf[i++]);
      fputs("e\n", fp), fflush(fp);
      first_buf.clear();
    }
    if (!(plot%(plots[0]*plots[1])) && to_pause && !data_only)
      if (pause > 0)
	if (fp == fp_) sleep(pause); else fprintf(fp, "pause %d\n", pause);
      else if (pause == -1) {
	const string mesg = plot ?
	  "Press <ENTER> to plot the next screen." :
	  "Press <ENTER> to plot the next frame.";
	if (fp == fp_) { if (isatty(STDIN_FILENO)) cerr << mesg, cin.get(); }
	else fprintf(fp, "pause -1 \"%s\"\n", mesg.c_str());
      } else if (pause < -1) exit(0);
  }
  new_plot = true;
  return *this;
}

gnuplot& flush(gnuplot& gp) { return gp.flush(); }

void gnuplot::insert_nl()
{
  if (!data_only) fprintf(fp, "\n");
}

gnuplot& gnuplot::operator<<(const string s) {
  if (!data_only) flush(false), fputs(s.c_str(), fp);
  return *this;
}

gnuplot& gnuplot::operator<<(const double x) {
  if (new_plot && !data_only) {
    if (first_number) {
      fputs("set nokey\nset xrange [] writeback\nset yrange [] writeback\n"
	    "set zrange [] writeback\n", fp);
      set_title(title);
      set_xlabel(xlabel), set_ylabel(ylabel), set_zlabel(zlabel);
      first_number = false;
    }
    if (multiplot) {
      if (++plot == 1)  // the first plot (top left corner)
	fputs(plot3d ? "set format ''\n"
	      "set tmargin 0\nset bmargin 0\nset rmargin 0\nset lmargin 0\n" :
	      "set tmargin 0\nset bmargin 1\nset rmargin 4\nset lmargin 0\n",
	      fp);
      if (plot == plots[0]*plots[1]+1)  // the last plot (bottom right corner)
#ifdef USE_ERRC
      	errc(0, 0, "%hhux%hhu grid exhausted, wrapping around starting with frame #%hu", plots[0], plots[1], plot);
#else

      	error(0, 0, "%hhux%hhu grid exhausted, wrapping around starting with frame #%hu", plots[0], plots[1], plot);
#endif
      if (!plot3d) {
	fputs("set format ''\n", fp);
	// if on the left boundry
	if (!((plot-1)%plots[0])) fputs("set format y\n", fp);
	// if on the bottom
	if (plot%(plots[0]*plots[1]) > plots[0]*plots[1]-plots[0]
	    || !(plot%(plots[0]*plots[1])))
	  fputs("set format x\n", fp);
      }
      // VERY dense and algorithmic but SEEMS to work [so don't touch]
      // if there is an easier way to do this, though, please let me know
      fprintf(fp, "set title '%g' ,%d\nset origin %g,%g\nclear\n", time,
	      plot3d ? 0 : -2,
	      (plot3d?1:0.9)*((plot%plots[0]?:plots[0])-1)/plots[0] +
	      (plot3d?0:0.1),
	      (plot3d?1:0.93) -
	      (plot3d?1:0.9)*((plot/plots[0]+(plot%plots[0]?1:0))%plots[1]?:plots[1])/plots[1]);
    }
    set_title(title);
    fprintf(fp, "%s '-' with %s 1", plot3d ? "splot" : "plot", style.c_str());
    for (typeof(graphs) i = 1; i < graphs; ++i)
      fprintf(fp, ", '-' with %s 1", style.c_str());
    putc('\n', fp);
    new_plot = false;
  }
  if (graph++/(plot3d?3:2)%graphs) buf[(graph-1)/(plot3d?3:2)-1].push_back(x);
  else {
    fprintf(fp, "%g%c", x, (n=++n%(plot3d?3:2)) ? ' ' : '\n');
    /*
     * Save the first dataset so that we can resend it again to circumvent some
     * really annoying Gnuplot bugs, namely that the first plot looks different
     * from all successive ones.
     */
    if (first_dataset && !data_only && graphs == 1) first_buf.push_back(x);
    if (graph == graphs*(plot3d?3:2)+1) graph = 1;
  }
  return *this;
}

gnuplot& gnuplot::operator<<(gnuplot& (*f)(gnuplot&)) { return f(*this); }

void gnuplot::reset_fp() {
  data_only = false;
  fp = fp_;
}

void gnuplot::set_fp(FILE* const f, const bool d) {
  data_only = d;
  fp = f;
}

void gnuplot::set_graphs(unsigned short n) { buf.resize((graphs=n) - 1); }

void gnuplot::set_log(const string s) { *this << "set logscale " << s << "\n"; }

void gnuplot::set_multiplot(const unsigned char x, const unsigned char y) {
  *this << "set multiplot\n";
  if (x*y > 1 && !data_only) {
    const float relsize = plot3d ? 1 : 0.9;
    multiplot = true;
    fprintf(fp, "set size %g,%g\n", relsize/(plots[0]=x), relsize/(plots[1]=y));
  }
}

void gnuplot::set_multiplot(const string grid) {
  unsigned char x, y;
  if (sscanf(grid.c_str(), "%hhux%hhu", &x, &y) != 2) y = x;
  set_multiplot(x,y);
}

void gnuplot::set_nolog(const string s) {
  *this << "set nologscale " << s << "\n";
}

void gnuplot::set_nomultiplot() {
  *this << "set nomultiplot\n";
  if (multiplot) multiplot = false, *this << "set size 1,1\nset origin 0,0\n";
}

void gnuplot::set_output(const string out) {
  *this << "set size square\nset terminal postscript portrait enhanced color\n";
  if (out != "-") *this << "set output '" << out << "'\n";
}

void gnuplot::set_title(const string title) {
  if (multiplot)
    fprintf(fp, "set label \"%s\" at screen 0.5,0.97 center\n", title.c_str());
  else if (time == INVALID_TIME)
    fprintf(fp, "set title \"%s\"\n", title.c_str());
  else fprintf(fp, "set title \"%s\\nt = %g\"\n", title.c_str(), time);
}

void gnuplot::set_xlabel(const string s) {
  if (!multiplot) *this << "set xlabel \"" << s << "\"\n";
}

void gnuplot::set_xrange(const string s) {
  *this << "set xrange [" << s << "]\n";
}

void gnuplot::set_ylabel(const string s) {
  if (!multiplot) *this << "set ylabel \"" << s << "\"\n";
}

void gnuplot::set_yrange(const string s) {
  *this << "set yrange [" << s << "]\n";
}

void gnuplot::set_zlabel(const string s) {
  if (!multiplot) *this << "set zlabel \"" << s << "\"\n";
}

void gnuplot::set_zrange(const string s) {
  *this << "set zrange [" << s << "]\n";
}
