#include "visit_module.h"
#include "visit_module_datatypes.h"

#define SAFE_STRDUP(ptr) (((ptr) != NULL) ? strdup(ptr) : NULL)

#define VISIT_COMMAND 1
#define VISIT_COMPLETE_CONN 2
#define VISIT_DISCONNECT 3


// declare some global vars
// these should be declared extern in the header
namespace visit_module {

    boost::thread *visit_thread;
    VisitCache cache;
}

using namespace visit_module;

void VisitCache::free_Mesh(Mesh &m) {

    free_MeshMD(m.md);
    free_MeshData(m.md->meshType, m.data);

}

void VisitCache::free_Variable(Variable &v) {

    free_ScalarMD(v.md);
    free_ScalarData(v.data);

}

void VisitCache::free_SimMD() {
    
    
    if(sim_md != 0) {
        
        // free all of the strings in each metadata struct
        for(int i = 0; i < sim_md->numMeshes; ++i) {
            free((void*)sim_md->meshes[i].name);
            free((void*)sim_md->meshes[i].blockTitle);
            free((void*)sim_md->meshes[i].blockPieceName);
            free((void*)sim_md->meshes[i].groupTitle);
            free((void*)sim_md->meshes[i].units);
            free((void*)sim_md->meshes[i].xLabel);
            free((void*)sim_md->meshes[i].yLabel);
            free((void*)sim_md->meshes[i].zLabel);
        }
        free(sim_md->meshes);
        
        for(int i = 0; i < sim_md->numScalars; ++i) {
            free((void*)sim_md->scalars[i].name);
            free((void*)sim_md->scalars[i].meshName);
        }
        free(sim_md->scalars);

        free(sim_md);
        sim_md = 0;
    }
}

void VisitCache::free_MeshMD(VisIt_MeshMetaData *mmd)
{
    // Free the mesh metadata.
    free((void*)mmd->name);
    free((void*)mmd->blockTitle);
    free((void*)mmd->blockPieceName);
    free((void*)mmd->groupTitle);
    free((void*)mmd->units);
    free((void*)mmd->xLabel);
    free((void*)mmd->yLabel);
    free((void*)mmd->zLabel);
    free(mmd);
}

VisIt_MeshMetaData* 
VisitCache::copy_MeshMD(const VisIt_MeshMetaData *mmd) {

    VisIt_MeshMetaData *newmd = (VisIt_MeshMetaData*) malloc(sizeof(VisIt_MeshMetaData));
    memcpy(newmd, mmd, sizeof(VisIt_MeshMetaData));

    newmd->name = SAFE_STRDUP(mmd->name);
    newmd->blockTitle = SAFE_STRDUP(mmd->blockTitle);
    newmd->blockPieceName = SAFE_STRDUP(mmd->blockPieceName);
    newmd->groupTitle = SAFE_STRDUP(mmd->groupTitle);
    newmd->units = SAFE_STRDUP(mmd->units);
    newmd->xLabel = SAFE_STRDUP(mmd->xLabel);
    newmd->yLabel = SAFE_STRDUP(mmd->yLabel);
    newmd->zLabel = SAFE_STRDUP(mmd->zLabel);

    return newmd;
}

void VisitCache::free_MeshData(int mesh_type, void *mesh) {
    
    if(mesh_type == VISIT_MESHTYPE_RECTILINEAR) {
        VisIt_RectilinearMesh *m = (VisIt_RectilinearMesh*)mesh;
        free_DataArray(m->xcoords);
        free_DataArray(m->ycoords);
        free_DataArray(m->zcoords);
    }
    else if(mesh_type == VISIT_MESHTYPE_POINT) {
        VisIt_PointMesh *m = (VisIt_PointMesh*)mesh;
        free_DataArray(m->xcoords);
        free_DataArray(m->ycoords);
        free_DataArray(m->zcoords);
    }
    free(mesh);
}


void VisitCache::free_ScalarMD(VisIt_ScalarMetaData *md) {

    free((void*)md->name);
    free((void*)md->meshName);

    free(md);
}

void VisitCache::free_ScalarData(VisIt_ScalarData *sd) {

    free_DataArray(sd->data);
    
    free(sd);
}

VisIt_ScalarMetaData* 
VisitCache::copy_ScalarMD(const VisIt_ScalarMetaData *md) {

    VisIt_ScalarMetaData *newmd = (VisIt_ScalarMetaData*) malloc(sizeof(VisIt_ScalarMetaData));
    memcpy(newmd, md, sizeof(VisIt_ScalarMetaData));

    newmd->name = SAFE_STRDUP(md->name);
    newmd->meshName = SAFE_STRDUP(md->meshName);
    
    return newmd;
}

void VisitCache::free_DataArray(VisIt_DataArray da) {
    
    if(da.owner != VISIT_OWNER_SIM)
        return;

    switch(da.dataType)
    {
        case VISIT_DATATYPE_CHAR:
            free(da.cArray);
            da.cArray = NULL;
            break;
        case VISIT_DATATYPE_INT:
            free(da.iArray);
            da.iArray = NULL;
            break;
        case VISIT_DATATYPE_FLOAT:
            free(da.iArray);
            da.iArray = NULL;
            break;
        case VISIT_DATATYPE_DOUBLE:
            free(da.dArray);
            da.dArray = NULL;
            break;
    }
}
    


VisitCache::VisitCache() 
    : _mutex(), mesh_map(), var_map() {
    
    sim_md = 0;
}

VisitCache::~VisitCache() {
    
    
    map<string, Mesh>::iterator m_it;
    map<string, Variable>::iterator v_it;

    for(m_it = mesh_map.begin(); m_it != mesh_map.end(); m_it++) {
        free_Mesh(m_it->second);
    }

    for(v_it = var_map.begin(); v_it != var_map.end(); v_it++) {
        free_Variable(v_it->second);
    }


    free_SimMD();
}

void VisitCache::make_point_mesh(string name, int nDims, Attribute *x, Attribute *y, Attribute *z) {

    boost::mutex::scoped_lock lock(_mutex);
    
    int local_low, local_high, local_size;
    
    if(mesh_map.find(name) != mesh_map.end())
        free_Mesh(mesh_map[name]);
    else
        mesh_map[name] = Mesh();

    Mesh &m = mesh_map[name];
    m.name = name;
    m.type = VISIT_MESHTYPE_POINT;
    m.nDims = nDims;

    x->get_gen_distribution(BASIN_ID, &local_low, &local_high);
    local_size = local_high - local_low + 1;
    
    m.md = (VisIt_MeshMetaData*) malloc(sizeof(VisIt_MeshMetaData));
    m.md->name = strdup(m.name.c_str());
    m.md->meshType = m.type;
    m.md->topologicalDimension = m.nDims;
    m.md->spatialDimension = m.nDims;
    m.md->numBlocks = BASIN_NPROCS;
    m.md->blockTitle = strdup("Domains");
    m.md->blockPieceName = strdup("domain");
    m.md->numGroups = 0;
    m.md->groupTitle = strdup("groups");
    m.md->groupPieceName = strdup("group");
    m.md->groupIds = NULL;
    m.md->units = NULL;
    m.md->xLabel = NULL;
    m.md->yLabel = NULL;
    m.md->zLabel = NULL;


    // begin building the mesh data
    m.data = (VisIt_MeshData*) malloc(sizeof(VisIt_MeshData));
    m.data->meshType = m.md->meshType;
    m.data->pmesh = (VisIt_PointMesh*) malloc(sizeof(VisIt_PointMesh));
    m.data->pmesh->ndims = m.nDims;
    m.data->pmesh->nnodes = local_size;
    
    Attribute_Data_Type t = x->get_data_type();

    if(t == BASIN_DATA_DOUBLE) {
        double *x_d = (double*) malloc(sizeof(double)*local_size); 
        double *y_d = (double*) malloc(sizeof(double)*local_size); 
        double *z_d = 0;
        if(m.nDims == 3)
            z_d = (double*) malloc(sizeof(double)*local_size); 
        else
            z_d = (double*) malloc(sizeof(double));

        x->get(local_low, local_high, x_d);
        y->get(local_low, local_high, y_d);
        if(m.nDims == 3)
            z->get(local_low, local_high, z_d);

        m.data->pmesh->xcoords = 
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, x_d);
        m.data->pmesh->ycoords = 
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, y_d);
        m.data->pmesh->zcoords = 
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, z_d);
    }
    else if(t == BASIN_DATA_FLOAT) {
        float *x_d = (float*) malloc(sizeof(float)*local_size); 
        float *y_d = (float*) malloc(sizeof(float)*local_size); 
        float *z_d = 0;
        if(m.nDims == 3)
            z_d = (float*) malloc(sizeof(float)*local_size); 
        else
            z_d = (float*) malloc(sizeof(float));

        x->get(local_low, local_high, x_d);
        y->get(local_low, local_high, y_d);
        if(m.nDims == 3)
            z->get(local_low, local_high, z_d);

        m.data->pmesh->xcoords = 
            VisIt_CreateDataArrayFromFloat(VISIT_OWNER_SIM, x_d);
        m.data->pmesh->ycoords = 
            VisIt_CreateDataArrayFromFloat(VISIT_OWNER_SIM, y_d);
        m.data->pmesh->zcoords = 
            VisIt_CreateDataArrayFromFloat(VISIT_OWNER_SIM, z_d);
    }
    else if(t == BASIN_DATA_INT) {
        int *x_d = (int*) malloc(sizeof(int)*local_size); 
        int *y_d = (int*) malloc(sizeof(int)*local_size); 
        int *z_d = 0;
        if(m.nDims == 3)
            z_d = (int*) malloc(sizeof(int)*local_size); 
        else
            z_d = (int*) malloc(sizeof(int));

        x->get(local_low, local_high, x_d);
        y->get(local_low, local_high, y_d);
        if(m.nDims == 3)
            z->get(local_low, local_high, z_d);

        m.data->pmesh->xcoords = 
            VisIt_CreateDataArrayFromInt(VISIT_OWNER_SIM, x_d);
        m.data->pmesh->ycoords = 
            VisIt_CreateDataArrayFromInt(VISIT_OWNER_SIM, y_d);
        m.data->pmesh->zcoords = 
            VisIt_CreateDataArrayFromInt(VISIT_OWNER_SIM, z_d);
    }
    else if(t == BASIN_DATA_CHAR) {    
        char *x_d = (char*) malloc(sizeof(char)*local_size); 
        char *y_d = (char*) malloc(sizeof(char)*local_size); 
        char *z_d = 0;
        if(m.nDims == 3)
            z_d = (char*) malloc(sizeof(char)*local_size); 
        else
            z_d = (char*) malloc(sizeof(char));

        x->get(local_low, local_high, x_d);
        y->get(local_low, local_high, y_d);
        if(m.nDims == 3)
            z->get(local_low, local_high, z_d);

        m.data->pmesh->xcoords = 
            VisIt_CreateDataArrayFromChar(VISIT_OWNER_SIM, x_d);
        m.data->pmesh->ycoords = 
            VisIt_CreateDataArrayFromChar(VISIT_OWNER_SIM, y_d);
        m.data->pmesh->zcoords = 
            VisIt_CreateDataArrayFromChar(VISIT_OWNER_SIM, z_d);
    }
    
}

void VisitCache::make_grid_mesh(string name, vector<double> &ll, vector<double> &ur, vector<int> &nx, int li, int hi) {

    boost::mutex::scoped_lock lock(_mutex);
    
    // we distribute along nx[0] so, the l_size contains the
    // number of rows of local data
    int l_size = hi - li + 1;

    //printf("BASINID %i: li = %i hi = %i\n", BASIN_ID, li, hi);
    if(mesh_map.find(name) != mesh_map.end())
        free_Mesh(mesh_map[name]);
    else
        mesh_map[name] = Mesh();

    Mesh &m = mesh_map[name];
    m.name = name;
    m.type = VISIT_MESHTYPE_RECTILINEAR;
    m.nDims = ll.size();

    m.md = (VisIt_MeshMetaData*) malloc(sizeof(VisIt_MeshMetaData));
    m.md->name = strdup(m.name.c_str());
    m.md->meshType = m.type;
    m.md->topologicalDimension = m.nDims;
    m.md->spatialDimension = m.nDims;
    m.md->numBlocks = BASIN_NPROCS;
    m.md->blockTitle = strdup("Domains");
    m.md->blockPieceName = strdup("domain");
    m.md->numGroups = 0;
    m.md->groupTitle = strdup("Groups");
    m.md->groupPieceName = strdup("group");
    m.md->groupIds = NULL;
    m.md->units = NULL;
    m.md->xLabel = NULL;
    m.md->yLabel = NULL;
    m.md->zLabel = NULL;

    //begin building the mesh data
    m.data = (VisIt_MeshData*) malloc(sizeof(VisIt_MeshData));
    m.data->meshType = m.md->meshType;
    m.data->rmesh = 
        (VisIt_RectilinearMesh*) malloc(sizeof(VisIt_RectilinearMesh));
    memset(m.data->rmesh, 0, sizeof(VisIt_RectilinearMesh));
    
    m.data->rmesh->ndims = m.nDims;
    m.data->rmesh->baseIndex[0] = 0;
    m.data->rmesh->baseIndex[1] = 0;
    m.data->rmesh->baseIndex[2] = 0;

    m.data->rmesh->minRealIndex[0] = 0;
    m.data->rmesh->minRealIndex[1] = 0;
    m.data->rmesh->minRealIndex[2] = 0;

    // visit plots with data stored like a[depth][columns][rows]
    // but we store as a[rows][columns][depth], so all of this
    // here needs to be backwards
    if(m.nDims == 2) {
        m.data->rmesh->dims[0] = nx[1] + 1;
        m.data->rmesh->dims[1] = l_size+1;
        m.data->rmesh->dims[2] = int(1);
        
        m.data->rmesh->maxRealIndex[0] = nx[1];
        m.data->rmesh->maxRealIndex[1] = hi+1;
        m.data->rmesh->maxRealIndex[2] = int(0);
    }
    else if(m.nDims == 3) {
        m.data->rmesh->dims[2] = l_size+ 1;
        m.data->rmesh->dims[1] = nx[1] + 1;
        m.data->rmesh->dims[0] = nx[2] + 1;
        
        m.data->rmesh->maxRealIndex[2] = hi+1;
        m.data->rmesh->maxRealIndex[1] = nx[1];
        m.data->rmesh->maxRealIndex[0] = nx[2];
    }

    
    double *rows_d = (double*) malloc(sizeof(double) * (l_size+1));
    double *columns_d = (double*) malloc(sizeof(double) * (nx[1]+1));
    double *depth_d = NULL;
    if(m.nDims == 3)
        depth_d = (double*) malloc(sizeof(double)* (nx[2]+1));
    else
        depth_d = (double*) malloc(sizeof(double));
  
    // fill the linear data arrays
    for(int i = li; i <= hi+1; i++) {
        rows_d[i-li] = (ur[0] - ll[0]) * double(i)/nx[0] + ll[0];
    }

    
    for(int i = 0; i <= nx[1]; i++) {
        columns_d[i] = (ur[1] - ll[1]) * double(i)/nx[1] + ll[1];
    }
    
    if(m.nDims == 3) {
        for(int i = 0; i <= nx[2]; i++) {
            depth_d[i] = (ur[2] - ll[2]) * double(i)/nx[2] + ll[2];
        }
    }
   
    if(m.nDims == 2) {
        m.data->rmesh->xcoords = 
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, columns_d);
        m.data->rmesh->ycoords =
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, rows_d);
        m.data->rmesh->zcoords =
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, depth_d);
    } 
    else if(m.nDims == 3) {
        m.data->rmesh->xcoords = 
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, depth_d);
        m.data->rmesh->ycoords =
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, columns_d);
        m.data->rmesh->zcoords =
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, rows_d);
    }

}
        

void VisitCache::make_variable(string name, string meshName, Attribute *d) {
    
    boost::mutex::scoped_lock lock(_mutex);
    
    // check that the mesh exists 
    if(mesh_map.find(meshName) != mesh_map.end())
    {
        Mesh *m = &mesh_map[meshName];
    }
    else {
        //TODO exceptions
        cerr << "Error! Named mesh does not exist!\n";
        return;
    }

    if(var_map.find(name) != var_map.end()) 
        free_Variable(var_map[name]);
    else
        var_map[name] = Variable();

    Variable &v = var_map[name];
    
    v.name = name;
    v.meshName = meshName;
    
    // build meta data

    v.md = (VisIt_ScalarMetaData*) malloc(sizeof(VisIt_ScalarMetaData));
    v.md->name = strdup(v.name.c_str());
    v.md->meshName = strdup(v.meshName.c_str());
    v.md->centering = VISIT_VARCENTERING_ZONE;
    v.md->treatAsASCII = 0;
    
    // build data
    //if(mesh_map[meshName].type == VISIT_MESHTYPE_POINT) {
        
    int local_low, local_high, local_row_size;

    d->get_gen_distribution(BASIN_ID, &local_low, &local_high);
    
    // row size is all that matters for 1d case
    local_row_size = local_high - local_low + 1;
    
    vector<int> dims = d->get_dims();

    // holds the local low and 
    // local high indexes for any dimensionality
    vector<int> low_v(dims.size(), 0);
    vector<int> high_v(dims.size(), 0);

    low_v[0] = local_low;
    high_v[0] = local_high;

    // v.size holds the linear size of the data
    v.size = local_row_size;

    // figure out how many elements for the rest of the dimensions
    for(int i = 1; i < dims.size(); i++) {
        v.size *= dims[i];
        high_v[i] = dims[i] - 1;    
    }
    
    //cerr << "v.size = " << v.size << "\n";
    //for(int i = 0; i < dims.size(); i++) {
    //    cerr << "BASIN_ID "<< BASIN_ID << ": low_v[" << i << "]=" << low_v[i] << "\n"
    //         << "BASIN_ID " << BASIN_ID << ": high_v[" << i << "]=" << high_v[i] << "\n";
    //}

    v.data = (VisIt_ScalarData*) malloc(sizeof(VisIt_ScalarData));
    v.data->len = v.size;
     
    Attribute_Data_Type t = d->get_data_type();
    
    if(t == BASIN_DATA_DOUBLE) { 
        double *loc_d = (double*) malloc(sizeof(double)*v.size);
        d->get(low_v, high_v, loc_d);
        v.data->data = 
            VisIt_CreateDataArrayFromDouble(VISIT_OWNER_SIM, loc_d);

        //for(int i = 0; i < v.size; i++) {
        //    printf("BASIN_ID %i: scalar_data[%i] = %f\n", BASIN_ID, i, loc_d[i]);
        //}
    }
    else if(t == BASIN_DATA_FLOAT) {
        float *loc_d = (float*) malloc(sizeof(float)*v.size);
        d->get(low_v, high_v, loc_d);
        v.data->data = 
            VisIt_CreateDataArrayFromFloat(VISIT_OWNER_SIM, loc_d);
    }
    else if(t == BASIN_DATA_INT) {
        int *loc_d = (int*) malloc(sizeof(int)*v.size);
        d->get(low_v, high_v, loc_d);
        v.data->data = 
            VisIt_CreateDataArrayFromInt(VISIT_OWNER_SIM, loc_d);
    }
    else if(t == BASIN_DATA_CHAR) {
        char *loc_d = (char*) malloc(sizeof(char)*v.size);
        d->get(low_v, high_v, loc_d);
        v.data->data = 
            VisIt_CreateDataArrayFromChar(VISIT_OWNER_SIM, loc_d);
    }
    //}
    //else if(mesh_map[meshName].type == VISIT_MESHTYPE_RECTILINEAR) {
        //TODO fill this in
        //It may be that this code could be made to be identical to the
        //case for a pointmesh. Only the local data should ever need to
        //be returned, so thats exactly what happens above
        //


    //}

}

// This method should always regenerate the simulation metadata
// based on what is currently in the maps because this is the only
// place that will update the meta data
VisIt_SimulationMetaData* VisitCache::get_md() {
    
    boost::mutex::scoped_lock lock(_mutex);
    
    free_SimMD();
    
    sim_md = (VisIt_SimulationMetaData*) malloc(sizeof(VisIt_SimulationMetaData));

    memset(sim_md, 0, sizeof(VisIt_SimulationMetaData));

    sim_md->currentMode = VISIT_SIMMODE_RUNNING;
    sim_md->currentCycle = 0;
    sim_md->currentTime = 0;

    // Account for meshes
    sim_md->numMeshes = mesh_map.size();

    sim_md->meshes = (VisIt_MeshMetaData*) malloc(sizeof(VisIt_MeshMetaData)* sim_md->numMeshes); //new VisIt_MeshMetaData[sim_md->numMeshes];
   
    map<string, Mesh>::iterator m_it = mesh_map.begin();
    for(int i = 0; i < sim_md->numMeshes; ++i) {
        // Do a deep copy of our stored metadata
        VisIt_MeshMetaData *newmd = copy_MeshMD(m_it->second.md);
        memcpy(&sim_md->meshes[i], newmd, sizeof(VisIt_MeshMetaData));
        free(newmd);
        
        //sim_md->meshes[i] = *m_it->second.md;
        m_it++;
    }

    // Account for scalars
    sim_md->numScalars = var_map.size();
    sim_md->scalars = new VisIt_ScalarMetaData[sim_md->numScalars];

    map<string, Variable>::iterator v_it = var_map.begin();
    for(int i = 0; i < sim_md->numScalars; ++i) {
        // Do a deep copy of our stored metadata
        VisIt_ScalarMetaData *newmd = copy_ScalarMD(v_it->second.md);
        memcpy(&sim_md->scalars[i], newmd, sizeof(VisIt_ScalarMetaData));
        free(newmd);

        //sim_md->scalars[i] = *v_it->second.md;
        v_it++;
    }
    
    sim_md->materials = 0;
    sim_md->curves = 0;
    sim_md->expressions = 0;
    sim_md->numMaterials = 0;
    sim_md->numCurves = 0;
    sim_md->numExpressions = 0;
    sim_md->genericCommands = 0;
    sim_md->customCommands = 0;
    
    return sim_md;
}

VisIt_MeshData* VisitCache::get_mesh(string name) {
    boost::mutex::scoped_lock lock(_mutex);
    
    return mesh_map[name].data;
}

VisIt_ScalarData* VisitCache::get_var(string name) {
    boost::mutex::scoped_lock lock(_mutex);
    
    return var_map[name].data;
}

/**
 * Utility functions for building VisIt command strings
 */

string build_pseudocolor_plot(string variable) {

    string command = "AddPlot(\"Pseudocolor\", \"" + variable + "\")\n" +
        "DrawPlots()\n";

    return command;
}

string build_volume_plot(string variable) {

    string command = "AddPlot(\"Volume\", \"" + variable + "\")\n" +
        "DrawPlots()\n";

    return command;
}


void visit_expose_discrete(Attribute& x, Attribute& y, Attribute& z, Attribute& scalar, string visit_name) {
    
    visit_module::cache.make_point_mesh(visit_name+"_mesh", 3, &x, &y, &z);
    visit_module::cache.make_variable(visit_name, visit_name+"_mesh", &scalar);
}

void visit_plot_discrete(Attribute& x, Attribute& y, Attribute& z, Attribute& scalar, string visit_name) {
    visit_expose_discrete(x, y, z, scalar, visit_name);

    VisItTimeStepChanged();

    string com = build_pseudocolor_plot(visit_name);

    VisItExecuteCommand(com.c_str());
}


void visit_expose_discrete(Attribute &x, Attribute& y, Attribute& scalar, string visit_name) {

    visit_module::cache.make_point_mesh(visit_name+"_mesh", 2, &x, &y, NULL);
    visit_module::cache.make_variable(visit_name, visit_name+"_mesh", &scalar);
}

void visit_plot_discrete(Attribute &x, Attribute& y, Attribute& scalar, string visit_name) {
    visit_expose_discrete(x, y, scalar, visit_name);

    VisItTimeStepChanged();

    string com = build_pseudocolor_plot(visit_name);

    VisItExecuteCommand(com.c_str());
}


void visit_expose_discrete(List &l, string x, string y, string z, string scalar, string visit_name) {
    
    visit_module::cache.make_point_mesh(visit_name+"_mesh", 3, l.get_attribute_pointer(x), l.get_attribute_pointer(y), l.get_attribute_pointer(z));
    visit_module::cache.make_variable(visit_name, visit_name+"_mesh", l.get_attribute_pointer(scalar));
}

void visit_plot_discrete(List &l, string x, string y, string z, string scalar, string visit_name) {
    visit_expose_discrete(l, x, y, z, scalar, visit_name);

    VisItTimeStepChanged();

    string com = build_pseudocolor_plot(visit_name);

    VisItExecuteCommand(com.c_str());
}


void visit_expose_discrete(List &l, string x, string y, string scalar, string visit_name) {
    
    visit_module::cache.make_point_mesh(visit_name+"_mesh", 2, l.get_attribute_pointer(x), l.get_attribute_pointer(y), NULL);
    visit_module::cache.make_variable(visit_name, visit_name+"_mesh", l.get_attribute_pointer(scalar));
}

void visit_plot_discrete(List &l, string x, string y, string scalar, string visit_name) {
    visit_expose_discrete(l, x, y, scalar, visit_name);

    VisItTimeStepChanged();

    string com = build_pseudocolor_plot(visit_name);

    VisItExecuteCommand(com.c_str());
}


void visit_expose_continuous(Grid& g, string attr_name, string visit_name) {

    Context *c = g.get_context_pointer();
    Attribute *a = g.get_attribute_pointer(attr_name);

    int li, hi;
    a->get_gen_distribution(BASIN_ID, &li, &hi);

    vector<double> ll = c->get_LowerCorner();
    vector<double> ur = c->get_UpperCorner();
    vector<int> nx = c->get_n_grid_points();

    visit_module::cache.make_grid_mesh(visit_name+"_mesh", ll, ur, nx, li, hi);
    visit_module::cache.make_variable(visit_name, visit_name+"_mesh", a);
    
}

void visit_plot_continuous(Grid& g, string attr_name, string visit_name) {
    
    visit_expose_continuous(g, attr_name, visit_name);

    VisItTimeStepChanged();

    string com = build_volume_plot(visit_name);

    VisItExecuteCommand(com.c_str());
}

void visit_expose_continuous(vector<double> ll, vector<double> ur, vector<int> nx, Attribute &a, string visit_name) {

    // Make sure argument dimensionalities match
    int nx_s = ll.size();
    if(ur.size() != nx_s || nx.size() != nx_s) {
        //TODO exceptions
        cerr << "visit_expose attribute dimensionality error! Attribute will not be plotted!\n";
        return;
    }


    int li, hi;

    a.get_gen_distribution(BASIN_ID, &li, &hi);
    
    // we use hi+1 since we need the bounds of the data indexes
    visit_module::cache.make_grid_mesh(visit_name+"_mesh", ll, ur, nx, li, hi);
    visit_module::cache.make_variable(visit_name, visit_name+"_mesh", &a);
}

void visit_plot_continuous(vector<double> ll, vector<double> ur, vector<int> nx, Attribute& a, string visit_name) {
    
    visit_expose_continuous(ll, ur, nx, a, visit_name);

    VisItTimeStepChanged();

    string com = build_volume_plot(visit_name);

    VisItExecuteCommand(com.c_str());
}

void visit_expose_continuous(Grid& g, Attribute& attr, string visit_name) {
    
    Context *c = g.get_context_pointer();
    
    int li, hi;
    attr.get_gen_distribution(BASIN_ID, &li, &hi);

    vector<double> ll = c->get_LowerCorner();
    vector<double> ur = c->get_UpperCorner();
    vector<int> nx = c->get_n_grid_points();

    visit_module::cache.make_grid_mesh(visit_name+"_mesh", ll, ur, nx, li, hi);
    visit_module::cache.make_variable(visit_name, visit_name+"_mesh", &attr);
}

void visit_plot_continuous(Grid& g, Attribute& attr, string visit_name) {
    
    visit_expose_continuous(g, attr, visit_name);

    VisItTimeStepChanged();

    string com = build_volume_plot(visit_name);

    VisItExecuteCommand(com.c_str());
}





// Code to implement the callbacks needed for VisIt

VisIt_SimulationMetaData* VisItGetMetaData(void) {

    return visit_module::cache.get_md();
}

VisIt_MeshData* VisItGetMesh(int domain, const char *name) {
    
    return visit_module::cache.get_mesh(string(name));
}

VisIt_ScalarData* VisItGetScalar(int domain, const char *name) {
    
    return visit_module::cache.get_var(string(name));
}

VisIt_DomainList *VisItGetDomainList() {
    int i;
    VisIt_DomainList *dl = new VisIt_DomainList;
    
    dl->nTotalDomains = BASIN_NPROCS;
    dl->nMyDomains = 1;
    dl->myDomains = VisIt_CreateDataArrayFromInt(VISIT_OWNER_VISIT, &BASIN_ID);

    return dl;
}

extern "C"
{
    VisIt_SimulationCallback visitCallbacks =
    {
        &VisItGetMetaData,
        &VisItGetMesh,
        NULL, /* material */
        NULL, /* species */
        &VisItGetScalar,
        NULL, /* curve */
        NULL, /* mixed scalar */
        &VisItGetDomainList
    };
}


// Code to implement the Visit thread loop

//MPI_Comm visit_comm;

static int visit_broadcast_int_callback(int *value, int sender)
{
    return MPI_Bcast(value, 1, MPI_INT, sender, visit_comm) ;
}

static int visit_broadcast_string_callback(char *str, int len, int sender)
{
    return MPI_Bcast(str, len, MPI_CHAR, sender, visit_comm) ;
}

int prep_vis_connection()
{
  MPI_Comm_dup(MPI_COMM_WORLD, &visit_comm);
  VisItSetupEnvironment();
  VisItSetBroadcastIntFunction(visit_broadcast_int_callback);
  VisItSetBroadcastStringFunction(visit_broadcast_string_callback);
  VisItSetParallel(true);
  VisItSetParallelRank(BASIN_ID);
  if (BASIN_ID==0)
    {
#ifdef BASIN_VISIT_OLD_CONTROL_HEADER 
        VisItInitializeSocketAndDumpSimFile("Basin.sim",
                      "Basin Visit connection ",
                      "/no/useful/path",
                      NULL,
                      NULL);
#endif
#ifdef BASIN_VISIT_NEW_CONTROL_HEADER
        VisItInitializeSocketAndDumpSimFile("Basin.sim",
                      "Basin Visit connection ",
                      "/no/useful/path",
                      NULL,
                      NULL,
                      NULL);
#endif
    }
  return 1;
}


void visit_module_start() {

    prep_vis_connection();

    // we don't want to start a new thread if one already exists
    if(!visit_thread) {
        visit_thread = new boost::thread(&visit_loop);
        cerr << "visit_module thread started on rank " << BASIN_ID << "\n";
    }
    else {
        cerr << "visit_module thread already exists on rank " << BASIN_ID << "\n";
    }

}

void visit_module_join() {
    
    if(visit_thread) {
        visit_thread->join();
        delete visit_thread;
        visit_thread = 0;
    }
}



void slaveVisItCommandCallback() {
    int command = VISIT_COMMAND;
    MPI_Bcast(&command, 1, MPI_INT, 0, visit_comm);
}


void visit_loop() {
    //Node 0 -> listens on socket
    if(BASIN_ID == 0)
    {
        int visitstate;
        while(true)
        {
            visitstate = VisItDetectInput(0, 0);
            if(visitstate >= -5 && visitstate <= -1) {
                    cerr << "Unrecoverable VisIt error!\n";
            }
            else if(visitstate == 1)
            {
                //cerr << "Visit is trying to connect!\n";
                int command = VISIT_COMPLETE_CONN;
                MPI_Bcast(&command, 1, MPI_INT, 0, visit_comm);
                MPI_Barrier(visit_comm);
                if(VisItAttemptToCompleteConnection()) {
                    MPI_Barrier(visit_comm);
                    //cerr << "VisIt connected!\n";
                    VisItSetSlaveProcessCallback(slaveVisItCommandCallback);
                }
                else {
                    cerr << "Error when VisIt tried to connect!\n";
                    cerr << "Node " << BASIN_ID << " Reported error: " << VisItGetLastError();
                }

            }
            else if(visitstate == 2)
            {
                //cerr << "Node0 needs to process visit command!\n";
                int success = VisItProcessEngineCommand();
                if(!success) {
                    int command = VISIT_DISCONNECT;
                    MPI_Bcast(&command, 1, MPI_INT, 0, visit_comm);
                    cerr << "Node " << BASIN_ID << " disconnecting from Visit!\n";
                    VisItDisconnect();
                }
            }

        }
    }
    else // All nodes > 0, listen to Node0 for what to do  
    {
        int command = 0;
        while(true) 
        {
            command = 0;
            MPI_Bcast(&command, 1, MPI_INT, 0, visit_comm);
            if(command == VISIT_COMMAND) {
                //cerr << "Node rank " << BASIN_ID << " got a visit command!\n";
                
                int success = VisItProcessEngineCommand();
                if(!success) {
                    cerr << "Node " << BASIN_ID << " disconnecting from Visit!\n";
                    VisItDisconnect();
                }
            }
            else if(command == VISIT_COMPLETE_CONN) {
                MPI_Barrier(visit_comm);
                if(VisItAttemptToCompleteConnection()) {
                    //cerr << "Visit connected to Node " << BASIN_ID << "\n";
                    MPI_Barrier(visit_comm);
                }
                else {
                    cerr << "ERROR! Visit connection error!\n";
                    cerr << "Node " << BASIN_ID << " Reported error: " << VisItGetLastError();
                }

            }
            else if(command == VISIT_DISCONNECT) {
                cerr << "Node " << BASIN_ID << " disconnecting from Visit!\n";
                VisItDisconnect();
            }

        }
    }

    cerr << "Exited thread on Node " << BASIN_ID << "\n";
}

